# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Image and tile manipulation (transforming, merging, etc).
"""
import io
from io import BytesIO


from PIL import Image, ImageChops
from PIL.TiffImagePlugin import ImageFileDirectory_v2, TiffTags
from mapproxy.image.opts import create_image, ImageFormat
from mapproxy.config import base_config
from mapproxy.srs import make_lin_transf, get_epsg_num

import logging
from functools import reduce
log = logging.getLogger('mapproxy.image')


magic_bytes = [
    ('png', (b"\211PNG\r\n\032\n",)),
    ('jpeg', (b"\xFF\xD8",)),
    ('tiff', (b"MM\x00\x2a", b"II\x2a\x00",)),
    ('gif', (b"GIF87a", b"GIF89a",)),
]


def peek_image_format(buf):
    buf.seek(0)
    header = buf.read(10)
    buf.seek(0)
    for format, bytes in magic_bytes:
        if header.startswith(bytes):
            return format
    return None


TIFF_MODELPIXELSCALETAG = 33550
TIFF_MODELTIEPOINTTAG = 33922
TIFF_GEOKEYDIRECTORYTAG = 34735


class GeoReference(object):
    def __init__(self, bbox, srs):
        self.bbox = bbox
        self.srs = srs

    def tiepoints(self):
        return (
            0.0, 0.0, 0.0,
            self.bbox[0], self.bbox[3], 0.0,
        )

    def pixelscale(self, img_size):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        return (
            float(width)/img_size[0], float(height)/img_size[1], 0.0,
        )

    def tiff_tags(self, img_size):
        tags = ImageFileDirectory_v2()
        tags[TIFF_MODELPIXELSCALETAG] = self.pixelscale(img_size)
        tags.tagtype[TIFF_MODELPIXELSCALETAG] = TiffTags.DOUBLE
        tags[TIFF_MODELTIEPOINTTAG] = self.tiepoints()
        tags.tagtype[TIFF_MODELTIEPOINTTAG] = TiffTags.DOUBLE

        model_type = 2 if self.srs.is_latlong else 1
        tags[TIFF_GEOKEYDIRECTORYTAG] = (
            1, 1, 0, 3,  # {KeyDirectoryVersion, KeyRevision, MinorRevision, NumberOfKeys}
            1024, 0, 1, model_type,  # 1 projected, 2 geographic (lat/long)
            1025, 0, 1, 1,  # 1 RasterIsArea, 2 RasterIsPoint
            3072, 0, 1, get_epsg_num(self.srs.srs_code),
        )
        tags.tagtype[TIFF_GEOKEYDIRECTORYTAG] = TiffTags.SHORT
        return tags


class BaseImageSource(object):
    """
    Virtual parent class for ImageSource and BlankImageSource
    """

    def __init__(self):
        raise Exception("Virtual class BaseImageSource, cannot be instanciated.")

    def as_image(self):
        raise Exception("Virtual class BaseImageSource, method as_image cannot be called.")

    def as_buffer(self, image_opts=None, format=None, seekable=False):
        raise Exception("Virtual class BaseImageSource, method as_buffer cannot be called.")

    def close_buffers(self):
        pass


class ImageSource(BaseImageSource):
    """
    This class wraps either a PIL image, a file-like object, or a file name.
    You can access the result as an image (`as_image` ) or a file-like buffer
    object (`as_buffer`).
    """

    def __init__(self, source, size=None, image_opts=None, cacheable=True, georef=None):
        """
        :param source: the image
        :type source: PIL `Image`, image file object, or filename
        :param format: the format of the ``source``
        :param size: the size of the ``source`` in pixel
        """
        self._img = None
        self._buf = None
        self._fname = None
        self.source = source
        self.image_opts = image_opts
        self._size = size
        self.cacheable = cacheable
        self.georef = georef
        self.authorize_stale = False

    @property
    def source(self):
        return self._img or self._buf or self._fname

    @source.setter
    def source(self, source):
        self._img = None
        self._buf = None
        if isinstance(source, str):
            self._fname = source
        elif isinstance(source, Image.Image):
            self._img = source
        else:
            self._buf = source

    def close_buffers(self):
        if self._buf:
            try:
                self._buf.close()
            except IOError:
                pass

    @property
    def filename(self):
        return self._fname

    def as_image(self):
        """
        Returns the image or the loaded image.

        :rtype: PIL `Image`
        """
        if not self._img:
            self._make_seekable_buf()
            log.debug('file(%s) -> image', self._fname or self._buf)

            try:
                img = Image.open(self._buf)
            except Exception:
                self.close_buffers()
                raise
            self._img = img
        if self.image_opts and self.image_opts.transparent and self._img.mode == 'P':
            self._img = self._img.convert('RGBA')
        return self._img

    def _make_seekable_buf(self):
        if not self._buf and self._fname:
            self._buf = open(self._fname, 'rb')
        else:
            try:
                self._buf.seek(0)
            except (io.UnsupportedOperation, AttributeError):
                # PIL needs file objects with seek
                self._buf = BytesIO(self._buf.read())

    def _make_readable_buf(self):
        if not self._buf and self._fname:
            self._buf = open(self._fname, 'rb')
        elif not hasattr(self._buf, 'seek'):
            if not isinstance(self._buf, ReadBufWrapper):
                self._buf = ReadBufWrapper(self._buf)
        else:
            try:
                self._buf.seek(0)
            except (io.UnsupportedOperation, AttributeError):
                # PIL needs file objects with seek
                self._buf = BytesIO(self._buf.read())

    def as_buffer(self, image_opts=None, format=None, seekable=False):
        """
        Returns the image as a file object.

        :param format: The format to encode an image.
                       Existing files will not be re-encoded.
        :rtype: file-like object
        """
        if format:
            image_opts = (image_opts or self.image_opts).copy()
            image_opts.format = ImageFormat(format)
        if not self._buf and not self._fname:
            if image_opts is None:
                image_opts = self.image_opts
            log.debug('image -> buf(%s)' % (image_opts.format,))
            self._buf = img_to_buf(self._img, image_opts=image_opts, georef=self.georef)
        else:
            self._make_seekable_buf() if seekable else self._make_readable_buf()
            if self.image_opts and image_opts and not self.image_opts.format and image_opts.format:
                # need actual image_opts.format for next check
                self.image_opts = self.image_opts.copy()
                self.image_opts.format = peek_image_format(self._buf)
            if self.image_opts and image_opts and self.image_opts.format != image_opts.format:
                log.debug('converting image from %s -> %s' % (self.image_opts, image_opts))
                self.source = self.as_image()
                self._buf = None
                self.image_opts = image_opts
                # hide fname to prevent as_buffer from reading the file
                fname = self._fname
                self._fname = None
                self.as_buffer(image_opts)
                self._fname = fname
        return self._buf

    @property
    def size(self):
        if self._size is None:
            self._size = self.as_image().size
        return self._size


def SubImageSource(source, size, offset, image_opts, cacheable=True):
    """
    Create a new ImageSource with `size` and `image_opts` and
    place `source` image at `offset`.
    """
    # force new image to contain alpha channel
    new_image_opts = image_opts.copy()
    new_image_opts.transparent = True
    img = create_image(size, new_image_opts)

    if not hasattr(source, 'as_image'):
        source = ImageSource(source)
    subimg = source.as_image()
    img.paste(subimg, offset)
    return ImageSource(img, size=size, image_opts=new_image_opts, cacheable=cacheable)


class BlankImageSource(BaseImageSource):
    """
    ImageSource for transparent or solid-color images.
    Implements optimized as_buffer() method.
    """

    def __init__(self, size, image_opts, cacheable=False):
        self.size = size
        self.image_opts = image_opts
        self._buf = None
        self._img = None
        self.cacheable = cacheable
        self.authorize_stale = False

    def as_image(self):
        if not self._img:
            self._img = create_image(self.size, self.image_opts)
        return self._img

    def as_buffer(self, image_opts=None, format=None, seekable=False):
        if not self._buf:
            image_opts = (image_opts or self.image_opts).copy()
            if format:
                image_opts.format = ImageFormat(format)
            image_opts.colors = 0
            self._buf = img_to_buf(self.as_image(), image_opts=image_opts)
        return self._buf

    def close_buffers(self):
        pass


class ReadBufWrapper(object):
    """
    This class wraps everything with a ``read`` method and adds support
    for ``seek``, etc. A call to everything but ``read`` will create a
    StringIO object of the ``readbuf``.
    """

    def __init__(self, readbuf):
        self.ok_to_seek = False
        self.readbuf = readbuf
        self.stringio = None

    def read(self, *args, **kw):
        if self.stringio:
            return self.stringio.read(*args, **kw)
        return self.readbuf.read(*args, **kw)

    def __iter__(self):
        if self.stringio:
            return iter(self.stringio)
        else:
            return iter(self.readbuf)

    def __getattr__(self, name):
        if self.stringio is None:
            if hasattr(self.readbuf, name):
                return getattr(self.readbuf, name)
            elif name == '__length_hint__':
                raise AttributeError
            self.ok_to_seek = True
            self.stringio = BytesIO(self.readbuf.read())
        return getattr(self.stringio, name)


def img_has_transparency(img):
    if img.mode == 'P':
        if img.info.get('transparency', False):
            return True
        # convert to RGBA and check alpha channel
        img = img.convert('RGBA')
    if img.mode == 'RGBA':
        # any alpha except fully opaque
        return any(img.histogram()[-256:-1])
    return False


def img_to_buf(img, image_opts, georef=None):
    defaults = {}
    image_opts = image_opts.copy()

    # convert I or L images to target mode
    if image_opts.mode and img.mode[0] in ('I', 'L') and img.mode != image_opts.mode:
        img = img.convert(image_opts.mode)

    if (image_opts.colors is None and base_config().image.paletted
            and image_opts.format.endswith('png')):
        # force 255 colors for png with globals.image.paletted
        image_opts.colors = 255

    format = filter_format(image_opts.format.ext)
    if format == 'mixed':
        if img_has_transparency(img):
            format = 'png'
            image_opts.transparent = True
        else:
            format = 'jpeg'
            image_opts.colors = None
            image_opts.transparent = False

    # quantize if colors is set, but not if we already have a paletted image
    if image_opts.colors and not (img.mode == 'P' and len(img.getpalette()) == image_opts.colors*3):
        quantizer = None
        if 'quantizer' in image_opts.encoding_options:
            quantizer = image_opts.encoding_options['quantizer']
        if image_opts.transparent:
            img = quantize(img, colors=image_opts.colors, alpha=True,
                           defaults=defaults, quantizer=quantizer)
        else:
            img = quantize(img, colors=image_opts.colors,
                           quantizer=quantizer)
        if hasattr(Image, 'RLE'):
            defaults['compress_type'] = Image.RLE

    buf = BytesIO()
    if format == 'jpeg':
        img = img.convert('RGB')
        if 'jpeg_quality' in image_opts.encoding_options:
            defaults['quality'] = image_opts.encoding_options['jpeg_quality']
        else:
            defaults['quality'] = base_config().image.jpeg_quality

    elif format == 'tiff':
        if georef:
            tags = georef.tiff_tags(img.size)
            defaults['tiffinfo'] = tags
        if 'tiff_compression' in image_opts.encoding_options:
            defaults['compression'] = image_opts.encoding_options['tiff_compression']
            if defaults['compression'] == 'jpeg':
                if 'jpeg_quality' in image_opts.encoding_options:
                    defaults['quality'] = image_opts.encoding_options['jpeg_quality']

    # unsupported transparency tuple can still be in non-RGB img.infos
    # see: https://github.com/python-pillow/Pillow/pull/2633
    if (format == 'png' and img.mode != 'RGB'
            and 'transparency' in img.info and isinstance(img.info['transparency'], tuple)):
        del img.info['transparency']

    img.save(buf, format, **defaults)
    buf.seek(0)
    return buf


def quantize(img, colors=256, alpha=False, defaults=None, quantizer=None):
    if hasattr(Image, 'FASTOCTREE') and quantizer in (None, 'fastoctree'):
        if not alpha:
            img = img.convert('RGB')
        try:
            if img.mode == 'P':
                # quantize with alpha does not work with P images
                img = img.convert('RGBA')
            img = img.quantize(colors, Image.FASTOCTREE)
        except ValueError:
            pass
    else:
        if alpha and img.mode == 'RGBA':
            img.load()  # split might fail if image is not loaded
            alpha = img.split()[3]
            img = img.convert('RGB').convert('P', palette=Image.ADAPTIVE, colors=colors-1)
            mask = Image.eval(alpha, lambda a: 255 if a <= 128 else 0)
            img.paste(255, mask)
            if defaults is not None:
                defaults['transparency'] = 255
        else:
            img = img.convert('RGB').convert('P', palette=Image.ADAPTIVE, colors=colors)

    return img


def filter_format(format):
    if format.lower() == 'geotiff':
        format = 'tiff'
    if format.lower().startswith('png'):
        format = 'png'
    return format


image_filter = {
    'nearest': Image.NEAREST,
    'bilinear': Image.BILINEAR,
    'bicubic': Image.BICUBIC
}


def is_single_color_image(image):
    """
    Checks if the `image` contains only one color.
    Returns ``False`` if it contains more than one color, else
    the color-tuple of the single color.
    """
    result = image.getcolors(1)
    # returns a list of (count, color), limit to one
    if result is None:
        return False

    color = result[0][1]
    if image.mode == 'P':
        palette = image.getpalette()
        return palette[color*3], palette[color*3+1], palette[color*3+2]

    return result[0][1]


def make_transparent(img, color, tolerance=10):
    """
    Create alpha channel for the given image and make each pixel
    in `color` full transparent.

    Returns an RGBA ImageSoruce.

    Modifies the image in-place, unless it needs to be converted
    first (P->RGB).

    :param color: RGB color tuple
    :param tolerance: tolerance applied to each color value
    """
    result = _make_transparent(img.as_image(), color, tolerance)
    image_opts = img.image_opts.copy()
    image_opts.transparent = True
    image_opts.mode = 'RGBA'
    return ImageSource(result, size=result.size, image_opts=image_opts)


def _make_transparent(img, color, tolerance=10):
    img.load()

    if img.mode == 'P':
        img = img.convert('RGBA')

    channels = img.split()
    mask_channels = []
    for ch, c in zip(channels, color):
        # create bit mask for each matched color
        low_c, high_c = c-tolerance, c+tolerance
        mask_channels.append(Image.eval(ch, lambda x: 255 if low_c <= x <= high_c else 0))

    # multiply channel bit masks to get a single mask
    alpha = reduce(ImageChops.multiply, mask_channels)
    # invert to get alpha channel
    alpha = ImageChops.invert(alpha)

    if len(channels) == 4:
        # multiply with existing alpha
        alpha = ImageChops.multiply(alpha, channels[-1])

    img.putalpha(alpha)
    return img


def bbox_position_in_image(bbox, size, src_bbox):
    """
    Calculate the position of ``bbox`` in an image of ``size`` and ``src_bbox``.
    Returns the sub-image size and the offset in pixel from top-left corner
    and the sub-bbox.

    >>> bbox_position_in_image((-180, -90, 180, 90), (600, 300), (-180, -90, 180, 90))
    ((600, 300), (0, 0), (-180, -90, 180, 90))
    >>> bbox_position_in_image((-200, -100, 200, 100), (600, 300), (-180, -90, 180, 90))
    ((540, 270), (30, 15), (-180, -90, 180, 90))
    >>> bbox_position_in_image((-200, -50, 200, 100), (600, 300), (-180, -90, 180, 90))
    ((540, 280), (30, 20), (-180, -50, 180, 90))
    >>> bbox_position_in_image((586400,196400,752800,362800), (256, 256), (586400,196400,752800,350000))
    ((256, 237), (0, 19), (586400, 196400, 752800, 350000))
    """
    coord_to_px = make_lin_transf(bbox, (0, 0) + size)
    offsets = [0, size[1], size[0], 0]
    sub_bbox = list(bbox)
    if src_bbox[0] > bbox[0]:
        sub_bbox[0] = src_bbox[0]
        x, y = coord_to_px((src_bbox[0], 0))
        offsets[0] = int(x)
    if src_bbox[1] > bbox[1]:
        sub_bbox[1] = src_bbox[1]
        x, y = coord_to_px((0, src_bbox[1]))
        offsets[1] = int(y)

    if src_bbox[2] < bbox[2]:
        sub_bbox[2] = src_bbox[2]
        x, y = coord_to_px((src_bbox[2], 0))
        offsets[2] = int(x)

    if src_bbox[3] < bbox[3]:
        sub_bbox[3] = src_bbox[3]
        x, y = coord_to_px((0, src_bbox[3]))
        offsets[3] = int(y)

    size = abs(offsets[2] - offsets[0]), abs(offsets[1] - offsets[3])
    return size, (offsets[0], offsets[3]), tuple(sub_bbox)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Service requests (parsing, handling, etc).
"""
from urllib.parse import parse_qsl, quote

from mapproxy.util.py import cached_property


class NoCaseMultiDict(dict):
    """
    This is a dictionary that allows case insensitive access to values.

    >>> d = NoCaseMultiDict([('A', 'b'), ('a', 'c'), ('B', 'f'), ('c', 'x'), ('c', 'y'), ('c', 'z')])
    >>> d['a']
    'b'
    >>> d.get_all('a')
    ['b', 'c']
    >>> 'a' in d and 'b' in d
    True
    """

    def _gen_dict(self, mapping=()):
        """A `NoCaseMultiDict` can be constructed from an iterable of
        ``(key, value)`` tuples or a dict.
        """
        tmp = {}
        if isinstance(mapping, NoCaseMultiDict):
            for key, value in mapping.iteritems():
                tmp.setdefault(key.lower(), (key, []))[1].extend(value)
        else:
            if isinstance(mapping, dict):
                itr = mapping.items()
            else:
                itr = iter(mapping)
            for key, value in itr:
                tmp.setdefault(key.lower(), (key, []))[1].append(value)
        return tmp

    def __init__(self, mapping=()):
        """A `NoCaseMultiDict` can be constructed from an iterable of
        ``(key, value)`` tuples or a dict.
        """
        dict.__init__(self, self._gen_dict(mapping))

    def update(self, mapping=(), append=False):
        """A `NoCaseMultiDict` can be updated from an iterable of
        ``(key, value)`` tuples or a dict.
        """
        for _, (key, values) in self._gen_dict(mapping).items():
            self.set(key, values, append=append, unpack=True)

    def __getitem__(self, key):
        """
        Return the first data value for this key.

        :raise KeyError: if the key does not exist
        """
        if key in self:
            return dict.__getitem__(self, key.lower())[1][0]
        raise KeyError(key)

    def __setitem__(self, key, value):
        dict.setdefault(self, key.lower(), (key, []))[1][:] = [value]

    def __delitem__(self, key):
        dict.__delitem__(self, key.lower())

    def __contains__(self, key):
        return dict.__contains__(self, key.lower())

    def __getstate__(self):
        data = []
        for key, values in self.iteritems():
            for v in values:
                data.append((key, v))
        return data

    def __setstate__(self, data):
        self.__init__(data)

    def get(self, key, default=None, type_func=None):
        """Return the default value if the requested data doesn't exist.
        If `type_func` is provided and is a callable it should convert the value,
        return it or raise a `ValueError` if that is not possible.  In this
        case the function will return the default as if the value was not
        found.

        Example:

        >>> d = NoCaseMultiDict(dict(foo='42', bar='blub'))
        >>> d.get('foo', type_func=int)
        42
        >>> d.get('bar', -1, type_func=int)
        -1
        """
        try:
            rv = self[key]
            if type_func is not None:
                rv = type_func(rv)
        except (KeyError, ValueError):
            rv = default
        return rv

    def get_all(self, key):
        """
        Return all values for the key as a list. Returns an empty list, if
        the key doesn't exist.
        """
        if key in self:
            return dict.__getitem__(self, key.lower())[1]
        else:
            return []

    def set(self, key, value, append=False, unpack=False):
        """
        Set a `value` for the `key`. If `append` is ``True`` the value will be added
        to other values for this `key`.

        If `unpack` is True, `value` will be unpacked and each item will be added.
        """
        if key in self:
            if not append:
                dict.__getitem__(self, key.lower())[1][:] = []
        else:
            dict.__setitem__(self, key.lower(), (key, []))
        if unpack:
            for v in value:
                dict.__getitem__(self, key.lower())[1].append(v)
        else:
            dict.__getitem__(self, key.lower())[1].append(value)

    def iteritems(self):
        """
        Iterates over all keys and values.
        """
        for _, (key, values) in dict.items(self):
            yield key, values

    def copy(self):
        """
        Returns a copy of this object.
        """
        return self.__class__(self)

    def __repr__(self):
        tmp = []
        for key, values in self.iteritems():
            tmp.append((key, values))
        return '%s(%r)' % (self.__class__.__name__, tmp)


def url_decode(qs, charset='utf-8', decode_keys=False, include_empty=True,
               errors='ignore'):
    """
    Parse query string `qs` and return a `NoCaseMultiDict`.
    """
    tmp = []
    for key, value in parse_qsl(qs, include_empty):
        if not isinstance(key, str):
            key = key.decode(charset, errors)
        if not isinstance(value, str):
            value = value.decode(charset, errors)
        tmp.append((key, value))
    return NoCaseMultiDict(tmp)


class Request(object):
    charset = 'utf8'

    def __init__(self, environ):
        self.environ = environ
        self.environ['mapproxy.request'] = self

        script_name = environ.get('HTTP_X_SCRIPT_NAME', '')
        if script_name:
            del environ['HTTP_X_SCRIPT_NAME']
            environ['SCRIPT_NAME'] = script_name
            path_info = environ.get('PATH_INFO', '')
            if path_info.startswith(script_name):
                environ['PATH_INFO'] = path_info[len(script_name):]

    @cached_property
    def args(self):
        if self.environ.get('QUERY_STRING'):
            return url_decode(self.environ['QUERY_STRING'], self.charset)
        else:
            return {}

    @property
    def path(self):
        path = self.environ.get('PATH_INFO', '')
        if path and isinstance(path, bytes):
            path = path.decode('utf-8')
        return path

    def pop_path(self):
        path = self.path.lstrip('/')
        if '/' in path:
            result, rest = path.split('/', 1)
            self.environ['PATH_INFO'] = '/' + rest
        else:
            self.environ['PATH_INFO'] = ''
            result = path
        if result:
            self.environ['SCRIPT_NAME'] = self.environ['SCRIPT_NAME'] + '/' + result
        return result

    @cached_property
    def host(self):
        if 'HTTP_X_FORWARDED_HOST' in self.environ:
            # might be a list, return first host only
            host = self.environ['HTTP_X_FORWARDED_HOST']
            host = host.split(',', 1)[0].strip()
            return host
        elif 'HTTP_HOST' in self.environ:
            host = self.environ['HTTP_HOST']
            if ':' in host:
                port = host.split(':')[1]
                if ((self.url_scheme, port) in (('https', '443'), ('http', '80'))):
                    host = host.split(':')[0]
            return host
        result = self.environ['SERVER_NAME']
        if ((self.url_scheme, self.environ['SERVER_PORT'])
                not in (('https', '443'), ('http', '80'))):
            result += ':' + self.environ['SERVER_PORT']
        return result

    @cached_property
    def url_scheme(self):
        scheme = self.environ.get('HTTP_X_FORWARDED_PROTO')
        if not scheme:
            scheme = self.environ['wsgi.url_scheme']
        return scheme

    @cached_property
    def accept_header(self):
        return self.environ.get('HTTP_ACCEPT', '')

    @cached_property
    def host_url(self):
        return '%s://%s/' % (self.url_scheme, self.host)

    @cached_property
    def server_url(self):
        return 'http://%s:%s/' % (
            self.environ['SERVER_NAME'],
            self.environ['SERVER_PORT']
        )

    @property
    def script_url(self):
        "Full script URL without trailing /"
        return (self.host_url.rstrip('/') +
                quote(self.environ.get('SCRIPT_NAME', '/').rstrip('/'))
                )

    @property
    def server_script_url(self):
        "Internal script URL"
        return self.server_url.rstrip('/')

    @property
    def base_url(self):
        return (self.host_url.rstrip('/')
                + quote(self.environ.get('SCRIPT_NAME', '').rstrip('/'))
                + quote(self.environ.get('PATH_INFO', ''))
                )


class RequestParams(object):
    """
    This class represents key-value request parameters. It allows case-insensitive
    access to all keys. Multiple values for a single key will be concatenated
    (eg. to ``layers=foo&layers=bar`` becomes ``layers: foo,bar``).

    All values can be accessed as a property.

    :param param: A dict or ``NoCaseMultiDict``.
    """
    params = None

    def __init__(self, param=None):
        self.delimiter = ','

        if param is None:
            self.params = NoCaseMultiDict()
        else:
            self.params = NoCaseMultiDict(param)

    def __str__(self):
        return self.query_string

    def get(self, key, default=None, type_func=None):
        """
        Returns the value for `key` or the `default`. `type_func` is called on the
        value to alter the value (e.g. use ``type_func=int`` to get ints).
        """
        return self.params.get(key, default, type_func)

    def set(self, key, value, append=False, unpack=False):
        """
        Set a `value` for the `key`. If `append` is ``True`` the value will be added
        to other values for this `key`.

        If `unpack` is True, `value` will be unpacked and each item will be added.
        """
        self.params.set(key, value, append=append, unpack=unpack)

    def update(self, mapping=(), append=False):
        """
        Update internal request parameters from an iterable of ``(key, value)``
        tuples or a dict.

        If `append` is ``True`` the value will be added to other values for
        this `key`.
        """
        self.params.update(mapping, append=append)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("'%s' object has no attribute '%s" %
                                 (self.__class__.__name__, name))

    def __getitem__(self, key):
        return self.delimiter.join(map(str, self.params.get_all(key)))

    def __setitem__(self, key, value):
        """
        Set `value` for the `key`. Does not append values (see ``MapRequest.set``).
        """
        self.set(key, value)

    def __delitem__(self, key):
        if key in self:
            del self.params[key]

    def iteritems(self):
        for key, values in self.params.iteritems():
            yield key, self.delimiter.join((str(x) for x in values))

    def __contains__(self, key):
        return self.params and key in self.params

    def copy(self):
        return self.__class__(self.params)

    @property
    def query_string(self):
        """
        The map request as a query string (the order is not guaranteed).

        >>> qs = RequestParams(dict(foo='egg', bar='ham%eggs', baz=100)).query_string
        >>> sorted(qs.split('&'))
        ['bar=ham%25eggs', 'baz=100', 'foo=egg']
        """
        kv_pairs = []
        for key, values in self.params.iteritems():
            value = ','.join(str(v) for v in values)
            kv_pairs.append(key + '=' + quote(value.encode('utf-8'), safe=','))
        return '&'.join(kv_pairs)

    def with_defaults(self, defaults):
        """
        Return this MapRequest with all values from `defaults` overwritten.
        """
        new = self.copy()
        for key, value in defaults.params.iteritems():
            if value != [None]:
                new.set(key, value, unpack=True)
        return new


class BaseRequest(object):
    """
    This class represents a request with a URL and key-value parameters.

    :param param: A dict, `NoCaseMultiDict` or ``RequestParams``.
    :param url: The service URL for the request.
    :param validate: True if the request should be validated after initialization.
    """
    request_params = RequestParams

    def __init__(self, param=None, url='', validate=False, http=None, dimensions=None):
        self.delimiter = ','
        self.http = http

        if param is None:
            self.params = self.request_params(NoCaseMultiDict())
        else:
            if isinstance(param, RequestParams):
                self.params = self.request_params(param.params)
            else:
                self.params = self.request_params(NoCaseMultiDict(param))
        self.url = url
        if validate:
            self.validate()

    def __str__(self):
        return self.complete_url

    def validate(self):
        pass

    @property
    def raw_params(self):
        params = {}
        for key, value in self.params.items():
            params[key] = value
        return params

    @property
    def query_string(self):
        return self.params.query_string

    @property
    def complete_url(self):
        """
        The complete MapRequest as URL.
        """
        if not self.url:
            return self.query_string
        delimiter = '?'
        if '?' in self.url:
            delimiter = '&'
        if self.url[-1] == '?':
            delimiter = ''
        return self.url + delimiter + self.query_string

    def copy_with_request_params(self, req):
        """
        Return a copy of this request ond overwrite all param values from `req`.
        Use this method for templates
        (``req_template.copy_with_request_params(actual_values)``).
        """
        new_params = req.params.with_defaults(self.params)
        return self.__class__(param=new_params, url=self.url)

    def __repr__(self):
        return '%s(param=%r, url=%r)' % (self.__class__.__name__, self.params, self.url)


def split_mime_type(mime_type):
    """
    >>> split_mime_type('text/xml; charset=utf-8')
    ('text', 'xml', 'charset=utf-8')
    """
    options = None
    mime_class = None
    if '/' in mime_type:
        mime_class, mime_type = mime_type.split('/', 1)
    if ';' in mime_type:
        mime_type, options = [part.strip() for part in mime_type.split(';', 2)]
    return mime_class, mime_type, options



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import functools

import pytest

from mapproxy.test.image import create_tmp_image
from mapproxy.test.http import MockServ
from mapproxy.test.helper import validate_with_xsd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "wmts_dimensions.yaml"


ns_wmts = {
    "wmts": "http://www.opengis.net/wmts/1.0",
    "ows": "http://www.opengis.net/ows/1.1",
    "xlink": "http://www.w3.org/1999/xlink",
}


def assert_xpath(xml, xpath, expected, namespaces=None):
    assert xml.xpath(xpath, namespaces=namespaces)[0] == expected


assert_xpath_wmts = functools.partial(assert_xpath, namespaces=ns_wmts)

DIMENSION_LAYER_BASE_REQ = (
    "/service1?styles=&format=image%2Fpng&height=256"
    "&bbox=-20037508.3428,0.0,0.0,20037508.3428"
    "&layers=foo,bar&service=WMS&srs=EPSG%3A900913"
    "&request=GetMap&width=256&version=1.1.1"
)
NO_DIMENSION_LAYER_BASE_REQ = DIMENSION_LAYER_BASE_REQ.replace(
    "/service1?", "/service2?"
)

WMTS_KVP_URL = (
    "/service?service=wmts&request=GetTile&version=1.0.0"
    "&tilematrixset=GLOBAL_MERCATOR&tilematrix=01&tilecol=0&tilerow=0&format=png&style="
)

TEST_TILE = create_tmp_image((256, 256))


class TestWMTS(SysTest):

    def test_capabilities(self, app):
        resp = app.get("/wmts/1.0.0/WMTSCapabilities.xml")
        xml = resp.lxml
        assert validate_with_xsd(
            xml, xsd_name="wmts/1.0/wmtsGetCapabilities_response.xsd"
        )

        assert len(xml.xpath("//wmts:Layer", namespaces=ns_wmts)) == 2
        assert (
            len(xml.xpath("//wmts:Contents/wmts:TileMatrixSet", namespaces=ns_wmts))
            == 1
        )

        assert set(
            xml.xpath(
                "//wmts:Contents/wmts:Layer/wmts:ResourceURL/@template",
                namespaces=ns_wmts,
            )
        ) == set(
            [
                "http://localhost/wmts/dimension_layer/{TileMatrixSet}/{Time}/{Elevation}/{TileMatrix}/{TileCol}/{TileRow}.png",  # noqa
                "http://localhost/wmts/no_dimension_layer/{TileMatrixSet}/{Time}/{Elevation}/{TileMatrix}/{TileCol}/{TileRow}.png",  # noqa
            ]
        )

        # check dimension values for dimension_layer
        dimension_elems = xml.xpath(
            '//wmts:Layer/ows:Identifier[text()="dimension_layer"]/following-sibling::wmts:Dimension',
            namespaces=ns_wmts,
        )
        dimensions = {}
        for elem in dimension_elems:
            dim = elem.find("{http://www.opengis.net/ows/1.1}Identifier").text
            default = elem.find("{http://www.opengis.net/wmts/1.0}Default").text
            values = [
                e.text for e in elem.findall("{http://www.opengis.net/wmts/1.0}Value")
            ]
            dimensions[dim] = (values, default)

        assert dimensions["Time"][0] == [
            "2012-11-12T00:00:00",
            "2012-11-13T00:00:00",
            "2012-11-14T00:00:00",
            "2012-11-15T00:00:00",
        ]
        assert dimensions["Time"][1] == "2012-11-15T00:00:00"
        assert dimensions["Elevation"][1] == "0"
        assert dimensions["Elevation"][0] == ["0", "1000", "3000"]

    def test_get_tile_valid_dimension(self, app):
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(
            DIMENSION_LAYER_BASE_REQ + "&Time=2012-11-15T00:00:00&elevation=1000"
        ).returns(TEST_TILE)
        with serv:
            resp = app.get(
                "/wmts/dimension_layer/GLOBAL_MERCATOR/2012-11-15T00:00:00/1000/01/0/0.png"
            )
        assert resp.content_type == "image/png"

    def test_get_tile_invalid_dimension(self, app):
        self.check_invalid_parameter(
            app,
            "/wmts/dimension_layer/GLOBAL_MERCATOR/2042-11-15T00:00:00/default/01/0/0.png",
        )

    def test_get_tile_default_dimension(self, app):
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(
            DIMENSION_LAYER_BASE_REQ + "&Time=2012-11-15T00:00:00&elevation=0"
        ).returns(TEST_TILE)
        with serv:
            resp = app.get(
                "/wmts/dimension_layer/GLOBAL_MERCATOR/default/default/01/0/0.png"
            )
        assert resp.content_type == "image/png"

    def test_get_tile_invalid_no_dimension_source(self, app):
        # unsupported dimension need to be 'default' in RESTful request
        self.check_invalid_parameter(
            app,
            "/wmts/no_dimension_layer/GLOBAL_MERCATOR/2042-11-15T00:00:00/default/01/0/0.png",
        )

    def test_get_tile_default_no_dimension_source(self, app):
        # check if dimensions are ignored
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(NO_DIMENSION_LAYER_BASE_REQ).returns(TEST_TILE)
        with serv:
            resp = app.get(
                "/wmts/no_dimension_layer/GLOBAL_MERCATOR/default/default/01/0/0.png"
            )
        assert resp.content_type == "image/png"

    def test_get_tile_kvp_valid_dimension(self, app):
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(
            DIMENSION_LAYER_BASE_REQ + "&Time=2012-11-14T00:00:00&elevation=3000"
        ).returns(TEST_TILE)
        with serv:
            resp = app.get(
                WMTS_KVP_URL
                + "&layer=dimension_layer&timE=2012-11-14T00:00:00&ELEvatioN=3000"
            )
        assert resp.content_type == "image/png"

    def test_get_tile_kvp_valid_dimension_defaults(self, app):
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(
            DIMENSION_LAYER_BASE_REQ + "&Time=2012-11-15T00:00:00&elevation=0"
        ).returns(TEST_TILE)
        with serv:
            resp = app.get(WMTS_KVP_URL + "&layer=dimension_layer")
        assert resp.content_type == "image/png"

    def test_get_tile_kvp_invalid_dimension(self, app):
        self.check_invalid_parameter(
            app, WMTS_KVP_URL + "&layer=dimension_layer&elevation=500"
        )

    def test_get_tile_kvp_default_no_dimension_source(self, app):
        # check if dimensions are ignored
        serv = MockServ(42423, bbox_aware_query_comparator=True)
        serv.expects(NO_DIMENSION_LAYER_BASE_REQ).returns(TEST_TILE)
        with serv:
            resp = app.get(
                WMTS_KVP_URL
                + "&layer=no_dimension_layer&Time=2012-11-14T00:00:00&Elevation=3000"
            )
        assert resp.content_type == "image/png"

    def check_invalid_parameter(self, app, url):
        resp = app.get(url, status=400)
        xml = resp.lxml
        assert resp.content_type == "text/xml"
        assert validate_with_xsd(xml, xsd_name="ows/1.1.0/owsExceptionReport.xsd")
        assert_xpath_wmts(
            xml,
            "/ows:ExceptionReport/ows:Exception/@exceptionCode",
            "InvalidParameterValue",
        )



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
import tempfile

import yaml

from mapproxy.script.conf.app import config_command
from mapproxy.test.helper import capture
from mapproxy.script.conf.geopackage import get_gpkg_contents, get_table_organization_coordsys_id, \
    get_table_tile_matrix, get_estimated_tile_res_ratio, get_res_table, get_geopackage_configuration_dict
from mapproxy.util.yaml import load_yaml


def filename(name):
    return os.path.join(os.path.dirname(__file__), "fixture", name)


class TestMapProxyConfCmd(object):

    def setup_method(self):
        self.dir = tempfile.mkdtemp()

    def teardown_method(self):
        if os.path.exists(self.dir):
            shutil.rmtree(self.dir)

    def tmp_filename(self, name):
        return os.path.join(self.dir, name)

    def get_test_gpkg(self):
        return os.path.join(os.path.join(os.path.dirname(__file__), 'fixture'), 'cache.gpkg')

    def test_cmd_no_args(self):
        with capture() as (stdout, stderr):
            assert config_command(["mapproxy-conf"]) == 2

        assert "--capabilities required" in stderr.getvalue()

    def test_stdout_output(self):
        with capture(bytes=True) as (stdout, stderr):
            assert (
                config_command(
                    [
                        "mapproxy-conf",
                        "--capabilities",
                        filename("util-conf-wms-111-cap.xml"),
                    ]
                )
                == 0
            )

        assert stdout.getvalue().startswith(b"# MapProxy configuration")

    def test_test_cap_output_no_base(self):
        with capture(bytes=True) as (stdout, stderr):
            assert (
                config_command(
                    [
                        "mapproxy-conf",
                        "--capabilities",
                        filename("util-conf-wms-111-cap.xml"),
                        "--output",
                        self.tmp_filename("mapproxy.yaml"),
                    ]
                )
                == 0
            )

        with open(self.tmp_filename("mapproxy.yaml"), "rb") as f:
            conf = yaml.safe_load(f)

            assert "grids" not in conf
            assert conf["sources"] == {
                "osm_roads_wms": {
                    "supported_srs": [
                        "CRS:84",
                        "EPSG:25831",
                        "EPSG:25832",
                        "EPSG:25833",
                        "EPSG:31466",
                        "EPSG:31467",
                        "EPSG:31468",
                        "EPSG:3857",
                        "EPSG:4258",
                        "EPSG:4326",
                        "EPSG:900913",
                    ],
                    "req": {
                        "layers": "osm_roads",
                        "url": "http://example.org/service",
                        "transparent": True,
                    },
                    "type": "wms",
                    "coverage": {
                        "srs": "EPSG:4326",
                        "bbox": [-180.0, -85.0511287798, 180.0, 85.0511287798],
                    },
                },
                "osm_wms": {
                    "supported_srs": [
                        "CRS:84",
                        "EPSG:25831",
                        "EPSG:25832",
                        "EPSG:25833",
                        "EPSG:31466",
                        "EPSG:31467",
                        "EPSG:31468",
                        "EPSG:3857",
                        "EPSG:4258",
                        "EPSG:4326",
                        "EPSG:900913",
                    ],
                    "req": {
                        "layers": "osm",
                        "url": "http://example.org/service",
                        "transparent": True,
                    },
                    "type": "wms",
                    "coverage": {
                        "srs": "EPSG:4326",
                        "bbox": [-180.0, -85.0511287798, 180.0, 85.0511287798],
                    },
                },
            }

            assert conf["layers"] == [
                {
                    "title": "ACME OpenStreetMap WMS",
                    "layers": [
                        {
                            "name": "osm",
                            "title": "OpenStreetMap (complete map)",
                            "sources": ["osm_wms"],
                        },
                        {
                            "name": "osm_roads",
                            "title": "OpenStreetMap (streets only)",
                            "sources": ["osm_roads_wms"],
                        },
                    ],
                }
            ]
            assert len(conf["layers"][0]["layers"]) == 2

    def test_test_cap_output(self):
        with capture(bytes=True) as (stdout, stderr):
            assert (
                config_command(
                    [
                        "mapproxy-conf",
                        "--capabilities",
                        filename("util-conf-wms-111-cap.xml"),
                        "--output",
                        self.tmp_filename("mapproxy.yaml"),
                        "--base",
                        filename("util-conf-base-grids.yaml"),
                    ]
                )
                == 0
            )

        with open(self.tmp_filename("mapproxy.yaml"), "rb") as f:
            conf = yaml.safe_load(f)

            assert "grids" not in conf
            assert len(conf["sources"]) == 2

            assert conf["caches"] == {
                "osm_cache": {
                    "grids": ["webmercator", "geodetic"],
                    "sources": ["osm_wms"],
                },
                "osm_roads_cache": {
                    "grids": ["webmercator", "geodetic"],
                    "sources": ["osm_roads_wms"],
                },
            }

            assert conf["layers"] == [
                {
                    "title": "ACME OpenStreetMap WMS",
                    "layers": [
                        {
                            "name": "osm",
                            "title": "OpenStreetMap (complete map)",
                            "sources": ["osm_cache"],
                        },
                        {
                            "name": "osm_roads",
                            "title": "OpenStreetMap (streets only)",
                            "sources": ["osm_roads_cache"],
                        },
                    ],
                }
            ]
            assert len(conf["layers"][0]["layers"]) == 2

    def test_test_gpkg_output(self):
        with capture(bytes=True) as (stdout, stderr):
            assert config_command(['mapproxy-conf',
                                   '--geopackage', filename('cache.gpkg'),
                                   '--output', self.tmp_filename('mapproxy.yaml'),
                                   '--base', filename('util-conf-base-grids.yaml'),
                                   ]) == 0

        with open(self.tmp_filename('mapproxy.yaml'), 'rb') as f:
            conf = load_yaml(f)

            assert conf['grids'] == {'cache_900913': {
                'origin': 'nw', 'srs': 'EPSG:900913',
                'res': [
                    156543.03392804097, 78271.51696402048, 39135.75848201024, 19567.87924100512,
                    9783.93962050256, 4891.96981025128, 2445.98490512564, 1222.99245256282,
                    611.49622628141, 305.748113140705, 152.8740565703525, 76.43702828517625,
                    38.21851414258813, 19.109257071294063, 9.554628535647032, 4.777314267823516,
                    2.388657133911758, 1.194328566955879, 0.5971642834779395],
                'bbox': [
                    -20037508.342789244, -20037508.342789244, 20037508.342789244,
                    20037508.342789244],
                'tile_size': [256, 256]}}

            conf['caches']['cache_cache']['cache']['filename'] = None

            assert conf['caches'] == {
                'cache_cache': {
                    'sources': [],
                    'cache': {'table_name': 'cache', 'type': 'geopackage', 'filename': None},
                    'grids': ['cache_900913']
                }
            }
            assert conf['layers'] == [{'sources': ['cache_cache'], 'name': 'cache', 'title': 'cache'}]

            assert conf['services'] == {'wms': None, 'demo': None, 'tms': {'origin': 'nw', 'use_grid_names': True},
                                        'kml': {'use_grid_names': True}, 'wmts': None}

    def test_overwrites(self):
        with capture(bytes=True) as (stdout, stderr):
            assert (
                config_command(
                    [
                        "mapproxy-conf",
                        "--capabilities",
                        filename("util-conf-wms-111-cap.xml"),
                        "--output",
                        self.tmp_filename("mapproxy.yaml"),
                        "--overwrite",
                        filename("util-conf-overwrite.yaml"),
                        "--base",
                        filename("util-conf-base-grids.yaml"),
                    ]
                )
                == 0
            )

        with open(self.tmp_filename("mapproxy.yaml"), "rb") as f:
            conf = yaml.safe_load(f)

            assert "grids" not in conf
            assert len(conf["sources"]) == 2

            assert conf["sources"] == {
                "osm_roads_wms": {
                    "supported_srs": ["EPSG:3857"],
                    "req": {
                        "layers": "osm_roads",
                        "url": "http://example.org/service",
                        "transparent": True,
                        "param": 42,
                    },
                    "type": "wms",
                    "coverage": {"srs": "EPSG:4326", "bbox": [0, 0, 90, 90]},
                },
                "osm_wms": {
                    "supported_srs": [
                        "CRS:84",
                        "EPSG:25831",
                        "EPSG:25832",
                        "EPSG:25833",
                        "EPSG:31466",
                        "EPSG:31467",
                        "EPSG:31468",
                        "EPSG:3857",
                        "EPSG:4258",
                        "EPSG:4326",
                        "EPSG:900913",
                    ],
                    "req": {
                        "layers": "osm",
                        "url": "http://example.org/service",
                        "transparent": True,
                        "param": 42,
                    },
                    "type": "wms",
                    "coverage": {
                        "srs": "EPSG:4326",
                        "bbox": [-180.0, -85.0511287798, 180.0, 85.0511287798],
                    },
                },
            }

            assert conf["caches"] == {
                "osm_cache": {
                    "grids": ["webmercator", "geodetic"],
                    "sources": ["osm_wms"],
                    "cache": {"type": "sqlite"},
                },
                "osm_roads_cache": {
                    "grids": ["webmercator"],
                    "sources": ["osm_roads_wms"],
                    "cache": {"type": "sqlite"},
                },
            }

            assert conf["layers"] == [
                {
                    "title": "ACME OpenStreetMap WMS",
                    "layers": [
                        {
                            "name": "osm",
                            "title": "OpenStreetMap (complete map)",
                            "sources": ["osm_cache"],
                        },
                        {
                            "name": "osm_roads",
                            "title": "OpenStreetMap (streets only)",
                            "sources": ["osm_roads_cache"],
                        },
                    ],
                }
            ]
            assert len(conf["layers"][0]["layers"]) == 2

    def test_get_gpkg_contents(self):
        returned_contents = get_gpkg_contents(self.get_test_gpkg())
        expected_results = [('cache', 'tiles', 'cache', 'Created with Mapproxy.', '2016-06-10T15:03:39.390Z',
                             -20037508.342789244, -20037508.342789244, 20037508.342789244, 20037508.342789244, 900913),
                            ('no_spatial_ref_sys', 'tiles', 'test_case', '', '2016-06-13T16:24:03.423Z',
                             -20037508.3427892, -20037508.3427892, 20037508.3427892, 20037508.3427892,
                             20037508.3427892)]
        assert expected_results == returned_contents

    def test_get_layer_organization_coordsys_id(self):
        returned_contents = get_table_organization_coordsys_id(self.get_test_gpkg(), 900913)
        expected_results = 900913
        assert expected_results == returned_contents

    def test_get_layer_organization_coordsys_id_2(self):
        returned_contents = get_table_tile_matrix(self.get_test_gpkg(), 'cache')
        expected_results = [(0, 1, 1, 256, 256, 156543.03392804097, 156543.03392804097),
                            (1, 2, 2, 256, 256, 78271.51696402048, 78271.51696402048),
                            (2, 4, 4, 256, 256, 39135.75848201024, 39135.75848201024),
                            (3, 8, 8, 256, 256, 19567.87924100512, 19567.87924100512),
                            (4, 16, 16, 256, 256, 9783.93962050256, 9783.93962050256),
                            (5, 32, 32, 256, 256, 4891.96981025128, 4891.96981025128),
                            (6, 64, 64, 256, 256, 2445.98490512564, 2445.98490512564),
                            (7, 128, 128, 256, 256, 1222.99245256282, 1222.99245256282),
                            (8, 256, 256, 256, 256, 611.49622628141, 611.49622628141),
                            (9, 512, 512, 256, 256, 305.748113140705, 305.748113140705),
                            (10, 1024, 1024, 256, 256, 152.8740565703525, 152.8740565703525),
                            (11, 2048, 2048, 256, 256, 76.43702828517625, 76.43702828517625),
                            (12, 4096, 4096, 256, 256, 38.21851414258813, 38.21851414258813),
                            (13, 8192, 8192, 256, 256, 19.109257071294063, 19.109257071294063),
                            (14, 16384, 16384, 256, 256, 9.554628535647032, 9.554628535647032),
                            (15, 32768, 32768, 256, 256, 4.777314267823516, 4.777314267823516),
                            (16, 65536, 65536, 256, 256, 2.388657133911758, 2.388657133911758),
                            (17, 131072, 131072, 256, 256, 1.194328566955879, 1.194328566955879),
                            (18, 262144, 262144, 256, 256, 0.5971642834779395, 0.5971642834779395)]
        assert expected_results == returned_contents

    def test_get_estimated_tile_res_ratio(self):
        # Test one level
        returned_contents = get_estimated_tile_res_ratio(((0, 1, 1, 256, 256, 156543.03392804097, 156543.03392804097),))
        expected_results = 2
        assert expected_results == returned_contents

        # Test two not contiguous levels
        returned_contents = get_estimated_tile_res_ratio(((0, 1, 1, 256, 256, 156543.03392804097, 156543.03392804097),
                                                          (2, 4, 4, 256, 256, 39135.75848201024, 39135.75848201024)))
        expected_results = 2
        assert expected_results == returned_contents

        # Test two contiguous levels
        returned_contents = get_estimated_tile_res_ratio(((0, 1, 1, 256, 256, 156543.03392804097, 156543.03392804097),
                                                          (1, 2, 2, 256, 256, 39135.75848201024, 39135.75848201024)))
        expected_results = 4
        assert expected_results == returned_contents

    def test_get_res_table(self):
        returned_contents = get_res_table([(5, 32, 32, 256, 256, 4891.96981025128, 4891.96981025128),
                                           (6, 64, 64, 256, 256, 2445.98490512564, 2445.98490512564)])
        expected_results = [156543.03392804097, 78271.51696402048, 39135.75848201024, 19567.87924100512,
                            9783.93962050256, 4891.96981025128, 2445.98490512564, 1222.99245256282, 611.49622628141,
                            305.748113140705, 152.8740565703525, 76.43702828517625, 38.21851414258813,
                            19.109257071294063, 9.554628535647032, 4.777314267823516, 2.388657133911758,
                            1.194328566955879, 0.5971642834779395]
        assert expected_results == returned_contents

    def test_get_geopackage_configuration_dict(self):
        returned_contents = get_geopackage_configuration_dict(self.get_test_gpkg())
        # See if filename is returned, but remove it so that test case can match regardless of install location.
        if returned_contents['caches']['cache_cache']['cache'].get('filename'):
            returned_contents['caches']['cache_cache']['cache']['filename'] = None
        expected_results = {'layers': [{'sources': ['cache_cache'], 'name': 'cache', 'title': 'cache'}],
                            'services': {'wms': None, 'demo': None, 'tms': {'origin': 'nw', 'use_grid_names': True},
                                         'kml': {'use_grid_names': True}, 'wmts': None}, 'grids': {
            'cache_900913': {'origin': 'nw', 'srs': 'EPSG:900913',
                             'res': [156543.03392804097, 78271.51696402048, 39135.75848201024, 19567.87924100512,
                                     9783.93962050256, 4891.96981025128, 2445.98490512564, 1222.99245256282,
                                     611.49622628141, 305.748113140705, 152.8740565703525, 76.43702828517625,
                                     38.21851414258813, 19.109257071294063, 9.554628535647032, 4.777314267823516,
                                     2.388657133911758, 1.194328566955879, 0.5971642834779395],
                             'bbox': [-20037508.342789244, -20037508.342789244, 20037508.342789244,
                                      20037508.342789244], 'tile_size': [256, 256]}}, 'caches': {
            'cache_cache': {'sources': [], 'cache': {'table_name': 'cache', 'type': 'geopackage',
                                                     'filename': None},
                            'grids': ['cache_900913']}}}
        assert expected_results == returned_contents



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division, print_function

import sys
import optparse
from itertools import cycle

DEFAULT_DPIS = {
    'OGC': 2.54/(0.00028 * 100),
}


def values_from_stdin():
    values = []
    for line in sys.stdin:
        line = line.split('#', 1)[0]
        if not line.strip():
            break
        values.append(float(line))
    return values


def scale_to_res(scale_denom, dpi, unit_factor):
    m_per_px = 2.54 / (dpi * 100)
    return scale_denom * m_per_px / unit_factor


def res_to_scale(res, dpi, unit_factor):
    m_per_px = 2.54 / (dpi * 100)
    return res / m_per_px * unit_factor


def format_simple(i, scale, res):
    return '%20.10f # %2d %20.8f' % (res, i, scale)


def format_list(i, scale, res):
    return '    %20.10f, # %2d %20.8f' % (res, i, scale)


def repeated_values(values, n):
    current_factor = 1
    step_factor = 10
    result = []
    for i, value in enumerate(slice(cycle(values), n)):
        if i != 0 and i % len(values) == 0:
            current_factor *= step_factor
        result.append(value/current_factor)
    return result


def fill_values(values, n):
    return values + [values[-1]/(2**x) for x in range(1, n)]


def scales_command(args=None):
    parser = optparse.OptionParser("%prog scales [options] scale/resolution[, ...]")
    parser.add_option("-l", "--levels", default=1, type=int, metavar='1',
                      help="number of resolutions/scales to calculate")
    parser.add_option("-d", "--dpi", default='OGC',
                      help="DPI to convert scales (use OGC for .28mm based DPI)")
    parser.add_option("--unit", default='m', metavar='m',
                      help="use resolutions in meter (m) or degrees (d)")
    parser.add_option("--eval", default=False, action='store_true',
                      help="evaluate args as Python expression. For example: 360/256")
    parser.add_option(
        "--repeat", default=False, action='store_true',
        help="repeat all values, each time /10. For example: 1000 500 250 results in 1000 500 250 100 50 25 10...")
    parser.add_option("--res-to-scale", default=False, action='store_true',
                      help="convert resolutions to scale")
    parser.add_option("--as-res-config", default=False, action='store_true',
                      help="output as resolution list for MapProxy grid configuration")

    if args:
        args = args[1:]  # remove script name
    (options, args) = parser.parse_args(args)
    options.levels = max(options.levels, len(args))

    dpi = float(DEFAULT_DPIS.get(options.dpi, options.dpi))

    if not args:
        parser.print_help()
        sys.exit(1)

    if args[0] == '-':
        values = values_from_stdin()
    elif options.eval:
        values = [eval(a) for a in args]
    else:
        values = [float(a) for a in args]

    values.sort(reverse=True)

    if options.repeat:
        values = repeated_values(values, options.levels)

    if len(values) < options.levels:
        values = fill_values(values, options.levels)

    unit_factor = 1
    if options.unit == 'd':
        # calculated from well-known scale set GoogleCRS84Quad
        unit_factor = 111319.4907932736

    calc = scale_to_res
    if options.res_to_scale:
        calc = res_to_scale

    if options.as_res_config:
        print('    res: [')
        print('         #  res            level     scale @%.1f DPI' % dpi)
        format = format_list
    else:
        format = format_simple

    for i, value in enumerate(values):
        print(format(i, value, calc(value, dpi, unit_factor)))

    if options.as_res_config:
        print('    ]')



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import shutil

from io import BytesIO

from mapproxy.request.wms import WMS111MapRequest
from mapproxy.test.http import MockServ
from mapproxy.test.image import is_png, create_tmp_image
from mapproxy.test.system import SysTest

import pytest


@pytest.fixture(scope="module")
def config_file():
    return "cache_mbtiles.yaml"


@pytest.fixture(scope="class")
def fixture_gpkg(base_dir):
    shutil.copy(
        os.path.join(os.path.dirname(__file__), "fixture", "cache.mbtiles"),
        base_dir.strpath,
    )


@pytest.mark.usefixtures("fixture_gpkg")
class TestMBTilesCache(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,-80,0,0",
                width="200",
                height="200",
                layers="mb",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )

    def test_get_map_cached(self, app):
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)

    def test_get_map_uncached(self, app):
        self.common_map_req.params.bbox = "-180,0,0,80"
        serv = MockServ(port=42423)
        serv.expects("/tiles/01/000/000/000/000/000/001.png")
        serv.returns(create_tmp_image((256, 256)))
        with serv:
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"
            data = BytesIO(resp.body)
            assert is_png(data)

        # now cached
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)



import math
from typing import Optional

from mapproxy.grid import (NoTiles, GridError, _create_tile_list, default_bboxs,
                           grid_bbox, origin_from_string)
from mapproxy.grid.resolutions import (aligned_resolutions, resolutions,
                                       pyramid_res_level, get_resolution,
                                       res_to_ogc_scale)
from mapproxy.srs import get_epsg_num, SRS, ogc_crs_url_to_auth_code
from mapproxy.util.bbox import bbox_equals, bbox_intersects, merge_bbox
from mapproxy.util.collections import ImmutableDictList

geodetic_epsg_codes = [4326]


class NamedGridList(ImmutableDictList):
    def __init__(self, items):
        tmp = []
        for i, value in enumerate(items):
            if isinstance(value, (tuple, list)):
                name, value = value
            else:
                name = str('%02d' % i)
            tmp.append((name, value))
        ImmutableDictList.__init__(self, tmp)


def tile_grid_for_epsg(epsg, bbox=None, tile_size=(256, 256), res=None):
    """
    Create a tile grid that matches the given epsg code:

    :param epsg: the epsg code
    :type epsg: 'EPSG:0000', '0000' or 0000
    :param bbox: the bbox of the grid
    :param tile_size: the size of each tile
    :param res: a list with all resolutions
    """
    epsg = get_epsg_num(epsg)
    if epsg in geodetic_epsg_codes:
        return TileGrid(epsg, is_geodetic=True, bbox=bbox, tile_size=tile_size, res=res)
    return TileGrid(epsg, bbox=bbox, tile_size=tile_size, res=res)


def tile_grid(srs=None, bbox=None, bbox_srs=None, tile_size=(256, 256),
              res=None, res_factor=2.0, threshold_res=None,
              num_levels=None, min_res=None, max_res=None,
              stretch_factor=1.15, max_shrink_factor=4.0,
              align_with=None, origin='ll', name=None
              ):
    """
    This function creates a new TileGrid.
    """
    if srs is None:
        srs = 'EPSG:900913'
    srs = SRS(srs)

    if not bbox:
        bbox = default_bboxs.get(srs)
        if not bbox:
            raise ValueError('need a bbox for grid with %s' % srs)

    bbox = grid_bbox(bbox, srs=srs, bbox_srs=bbox_srs)

    if res:
        if isinstance(res, list):
            if isinstance(res[0], (tuple, list)):
                # named resolutions
                res = sorted(res, key=lambda x: x[1], reverse=True)
            else:
                res = sorted(res, reverse=True)
            assert min_res is None
            assert max_res is None
            assert align_with is None
        else:
            raise ValueError("res is not a list, use res_factor for float values")

    elif align_with is not None:
        res = aligned_resolutions(min_res, max_res, res_factor, num_levels, bbox, tile_size,
                                  align_with)
    else:
        res = resolutions(min_res, max_res, res_factor, num_levels, bbox, tile_size)

    origin = origin_from_string(origin)

    return TileGrid(srs, bbox=bbox, tile_size=tile_size, res=res, threshold_res=threshold_res,
                    stretch_factor=stretch_factor, max_shrink_factor=max_shrink_factor,
                    origin=origin, name=name)


class UnsupportedException(Exception):
    def __init__(self, arg):
        Exception.__init__(self, arg)


def tile_grid_from_ogc_tile_matrix_set(ogc_tile_matrix_set):
    """ Return a TileGrid instance from the JSON representation of a
        OGC Two Dimensional Tile Matrix Set (https://docs.ogc.org/is/17-083r4/17-083r4.html)
    """

    srs = ogc_tile_matrix_set["crs"]
    try:
        srs = SRS(ogc_crs_url_to_auth_code(srs))
    except Exception:
        raise UnsupportedException(f"CRS {srs} is not supported")
    tileMatrices = ogc_tile_matrix_set["tileMatrices"]
    tileWidth = None
    tileHeight = None
    cornerOfOrigin = None
    origin = None
    bbox = None
    res = []
    grid_sizes = []
    for tileMatrix in tileMatrices:
        if "variableMatrixWidths" in tileMatrix:
            raise UnsupportedException("Tile matrix with variableMatrixWidths are not supported")

        cellSize = tileMatrix["cellSize"]
        matrixWidth = tileMatrix["matrixWidth"]
        matrixHeight = tileMatrix["matrixHeight"]

        if tileWidth is None:
            tileWidth = tileMatrix["tileWidth"]
            tileHeight = tileMatrix["tileHeight"]
            pointOfOrigin = tileMatrix["pointOfOrigin"]
            if srs.is_axis_order_ne:
                x = pointOfOrigin[1]
                y = pointOfOrigin[0]
            else:
                x = pointOfOrigin[0]
                y = pointOfOrigin[1]

            cornerOfOrigin = tileMatrix["cornerOfOrigin"]
            if cornerOfOrigin == "topLeft":
                origin = 'ul'
                bbox = [x,
                        y - tileHeight * matrixHeight * cellSize,
                        x + tileWidth * matrixWidth * cellSize,
                        y]
            elif cornerOfOrigin == "bottomLeft":
                origin = 'll'
                bbox = [x, y,
                        x + tileWidth * matrixWidth * cellSize,
                        y + tileHeight * matrixHeight * cellSize]
            else:
                raise UnsupportedException(f"Invalid cornerOfOrigin={cornerOfOrigin}")

        else:
            if tileWidth != tileMatrix["tileWidth"]:
                raise UnsupportedException("Tile matrix set with varying tileWidth depending on tile matrix")
            if tileHeight != tileMatrix["tileHeight"]:
                raise UnsupportedException("Tile matrix set with varying tileHeight depending on tile matrix")
            if pointOfOrigin != tileMatrix["pointOfOrigin"]:
                raise UnsupportedException("Tile matrix set with varying pointOfOrigin depending on tile matrix")
            if cornerOfOrigin != tileMatrix["cornerOfOrigin"]:
                raise UnsupportedException("Tile matrix set with varying cornerOfOrigin depending on tile matrix")

        res.append(cellSize)
        grid_sizes.append((matrixWidth, matrixHeight))

    tilegrid = TileGrid(srs=srs, bbox=bbox, tile_size=(tileWidth, tileHeight),
                    res=res, is_geodetic=srs.is_latlong, origin=origin,
                    name=ogc_tile_matrix_set["id"])

    grids = []
    idx = 0
    for name, _ in tilegrid.resolutions.iteritems():
        grids.append((name, grid_sizes[idx]))
        idx += 1
    tilegrid.grid_sizes = NamedGridList(grids)

    return tilegrid


class TileGrid(object):
    """
    This class represents a regular tile grid. The first level (0) contains a single
    tile, the origin is bottom-left.

    :ivar levels: the number of levels
    :ivar tile_size: the size of each tile in pixel
    :type tile_size: ``int(with), int(height)``
    :ivar srs: the srs of the grid
    :type srs: `SRS`
    :ivar bbox: the bbox of the grid, tiles may overlap this bbox
    """

    spheroid_a = 6378137.0  # for 900913
    flipped_y_axis = False

    def __init__(self, srs=900913, bbox=None, tile_size=(256, 256), res=None,
                 threshold_res=None, is_geodetic=False, levels=None,
                 stretch_factor=1.15, max_shrink_factor=4.0, origin='ll',
                 name=None):
        """
        :param stretch_factor: allow images to be scaled up by this factor
            before the next level will be selected
        :param max_shrink_factor: allow images to be scaled down by this
            factor before NoTiles is raised

        >>> grid = TileGrid(srs=900913)
        >>> [round(x, 2) for x in grid.bbox]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        """
        if isinstance(srs, (int, str)):
            srs = SRS(srs)
        self.srs = srs
        self.tile_size = tile_size
        self.origin = origin_from_string(origin)
        self.name = name

        if self.origin == 'ul':
            self.flipped_y_axis = True

        self.is_geodetic = is_geodetic

        self.stretch_factor = stretch_factor
        self.max_shrink_factor = max_shrink_factor

        if levels is None:
            self.levels = 20
        else:
            self.levels = levels

        if bbox is None:
            bbox = self._calc_bbox()
        self.bbox = bbox

        factor = None

        if res is None:
            factor = 2.0
            res = self._calc_res(factor=factor)
        elif res == 'sqrt2':
            if levels is None:
                self.levels = 40
            factor = math.sqrt(2)
            res = self._calc_res(factor=factor)
        elif is_float(res):
            factor = float(res)
            res = self._calc_res(factor=factor)

        self.levels = len(res)
        self.resolutions = NamedGridList(res)

        self.threshold_res = None
        if threshold_res:
            self.threshold_res = sorted(threshold_res)

        self.grid_sizes = self._calc_grids()

    def _calc_grids(self):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        grids = []
        for idx, res in self.resolutions.iteritems():
            x = max(math.ceil(width // res / self.tile_size[0]), 1)
            y = max(math.ceil(height // res / self.tile_size[1]), 1)
            grids.append((idx, (int(x), int(y))))
        return NamedGridList(grids)

    def _calc_bbox(self):
        if self.is_geodetic:
            return (-180.0, -90.0, 180.0, 90.0)
        else:
            circum = 2 * math.pi * self.spheroid_a
            offset = circum / 2.0
            return (-offset, -offset, offset, offset)

    def _calc_res(self, factor=None):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        initial_res = max(width/self.tile_size[0], height/self.tile_size[1])
        if factor is None:
            return pyramid_res_level(initial_res, levels=self.levels)
        else:
            return pyramid_res_level(initial_res, factor, levels=self.levels)

    def resolution(self, level):
        """
        Returns the resolution of the `level` in units/pixel.

        :param level: the zoom level index (zero is top)

        >>> grid = TileGrid(SRS(900913))
        >>> '%.5f' % grid.resolution(0)
        '156543.03393'
        >>> '%.5f' % grid.resolution(1)
        '78271.51696'
        >>> '%.5f' % grid.resolution(4)
        '9783.93962'
        """
        return self.resolutions[level]

    def closest_level(self, res):
        """
        Returns the level index that offers the required resolution.

        :param res: the required resolution
        :returns: the level with the requested or higher resolution

        >>> grid = TileGrid(SRS(900913))
        >>> grid.stretch_factor = 1.1
        >>> l1_res = grid.resolution(1)
        >>> [grid.closest_level(x) for x in (320000.0, 160000.0, l1_res+50, l1_res, \
                                             l1_res-50, l1_res*0.91, l1_res*0.89, 8000.0)]
        [0, 0, 1, 1, 1, 1, 2, 5]
        """
        prev_l_res = self.resolutions[0]
        threshold = None
        thresholds = []
        if self.threshold_res:
            thresholds = self.threshold_res[:]
            threshold = thresholds.pop()
            # skip thresholds above first res
            while threshold > prev_l_res and thresholds:
                threshold = thresholds.pop()

        threshold_result = None
        for level, l_res in enumerate(self.resolutions):
            if threshold and prev_l_res > threshold >= l_res:
                if res > threshold:
                    return level-1
                elif res >= l_res:
                    return level
                threshold = thresholds.pop() if thresholds else None

            if threshold_result is not None:
                # Use previous level that was within stretch_factor,
                # but only if this level res is smaller then res.
                # This fixes selection for resolutions that are closer together then stretch_factor.
                #
                if l_res < res:
                    return threshold_result

            if l_res <= res*self.stretch_factor:
                # l_res within stretch_factor
                # remember this level, check for thresholds or better res in next loop
                threshold_result = level
            prev_l_res = l_res
        return level

    def tile(self, x, y, level):
        """
        Returns the tile id for the given point.

        >>> grid = TileGrid(SRS(900913))
        >>> grid.tile(1000, 1000, 0)
        (0, 0, 0)
        >>> grid.tile(1000, 1000, 1)
        (1, 1, 1)
        >>> grid = TileGrid(SRS(900913), tile_size=(512, 512))
        >>> grid.tile(1000, 1000, 2)
        (2, 2, 2)
        """
        res = self.resolution(level)
        x = x - self.bbox[0]
        if self.flipped_y_axis:
            y = self.bbox[3] - y
        else:
            y = y - self.bbox[1]
        tile_x = x/float(res*self.tile_size[0])
        tile_y = y/float(res*self.tile_size[1])
        return (int(math.floor(tile_x)), int(math.floor(tile_y)), level)

    def flip_tile_coord(self, tile_coord):
        """
        Flip the tile coord on the y-axis. (Switch between bottom-left and top-left
        origin.)

        >>> grid = TileGrid(SRS(900913))
        >>> grid.flip_tile_coord((0, 1, 1))
        (0, 0, 1)
        >>> grid.flip_tile_coord((1, 3, 2))
        (1, 0, 2)
        """
        (x, y, z) = tile_coord
        return (x, self.grid_sizes[z][1]-1-y, z)

    def supports_access_with_origin(self, origin):
        if origin_from_string(origin) == self.origin:
            return True

        # check for each level if the top and bottom coordinates of the tiles
        # match the bbox of the grid. only in this case we can flip y-axis
        # without any issues

        # allow for some rounding errors in the _tiles_bbox calculations
        delta = max(abs(self.bbox[1]), abs(self.bbox[3])) / 1e12

        for level, grid_size in enumerate(self.grid_sizes):
            level_bbox = self._tiles_bbox([(0, 0, level),
                                           (grid_size[0] - 1, grid_size[1] - 1, level)])

            if abs(self.bbox[1] - level_bbox[1]) > delta or abs(self.bbox[3] - level_bbox[3]) > delta:
                return False
        return True

    def origin_tile(self, level, origin):
        assert self.supports_access_with_origin(origin), 'tile origins are incompatible'
        tile = (0, 0, level)

        if origin_from_string(origin) == self.origin:
            return tile

        return self.flip_tile_coord(tile)

    def get_affected_tiles(self, bbox, size, req_srs=None):
        """
        Get a list with all affected tiles for a bbox and output size.

        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> grid = TileGrid()
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> tile_size = (256, 256)
        >>> grid.get_affected_tiles(bbox, tile_size)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """
        src_bbox, level = self.get_affected_bbox_and_level(bbox, size, req_srs=req_srs)
        return self.get_affected_level_tiles(src_bbox, level)

    def get_affected_bbox_and_level(self, bbox, size, req_srs=None):
        if req_srs and req_srs != self.srs:
            src_bbox = req_srs.transform_bbox_to(self.srs, bbox)
        else:
            src_bbox = bbox

        if not bbox_intersects(self.bbox, src_bbox):
            raise NoTiles()

        res = get_resolution(src_bbox, size)
        level = self.closest_level(res)

        if res > self.resolutions[0]*self.max_shrink_factor:
            raise NoTiles()

        return src_bbox, level

    def get_affected_level_tiles(self, bbox, level):
        """
        Get a list with all affected tiles for a `bbox` in the given `level`.
        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> grid = TileGrid()
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> grid.get_affected_level_tiles(bbox, 0)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """
        # remove 1/10 of a pixel so we don't get a tiles we only touch
        delta = self.resolutions[level] / 10.0
        x0, y0, _ = self.tile(bbox[0]+delta, bbox[1]+delta, level)
        x1, y1, _ = self.tile(bbox[2]-delta, bbox[3]-delta, level)
        try:
            return self._tile_iter(x0, y0, x1, y1, level)
        except IndexError:
            raise GridError('Invalid BBOX')

    def _tile_iter(self, x0, y0, x1, y1, level):
        xs = list(range(x0, x1+1))
        if self.flipped_y_axis:
            y0, y1 = y1, y0
            ys = list(range(y0, y1+1))
        else:
            ys = list(range(y1, y0-1, -1))

        ll = (xs[0], ys[-1], level)
        ur = (xs[-1], ys[0], level)

        abbox = self._tiles_bbox([ll, ur])
        return (abbox, (len(xs), len(ys)),
                _create_tile_list(xs, ys, level, self.grid_sizes[level]))

    def _tiles_bbox(self, tiles):
        """
        Returns the bbox of multiple tiles.
        The tiles should be ordered row-wise, bottom-up.

        :param tiles: ordered list of tiles
        :returns: the bbox of all tiles
        """
        ll_bbox = self.tile_bbox(tiles[0])
        ur_bbox = self.tile_bbox(tiles[-1])
        return merge_bbox(ll_bbox, ur_bbox)

    def tile_bbox(self, tile_coord, limit=False):
        """
        Returns the bbox of the given tile.

        >>> grid = TileGrid(SRS(900913))
        >>> [round(x, 2) for x in grid.tile_bbox((0, 0, 0))]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        >>> [round(x, 2) for x in grid.tile_bbox((1, 1, 1))]
        [0.0, 0.0, 20037508.34, 20037508.34]
        """
        x, y, z = tile_coord
        res = self.resolution(z)

        x0 = self.bbox[0] + round(x * res * self.tile_size[0], 12)
        x1 = x0 + round(res * self.tile_size[0], 12)

        if self.flipped_y_axis:
            y1 = self.bbox[3] - round(y * res * self.tile_size[1], 12)
            y0 = y1 - round(res * self.tile_size[1], 12)
        else:
            y0 = self.bbox[1] + round(y * res * self.tile_size[1], 12)
            y1 = y0 + round(res * self.tile_size[1], 12)

        if limit:
            return (
                max(x0, self.bbox[0]),
                max(y0, self.bbox[1]),
                min(x1, self.bbox[2]),
                min(y1, self.bbox[3])
            )

        return x0, y0, x1, y1

    def limit_tile(self, tile_coord):
        """
        Check if the `tile_coord` is in the grid.

        :returns: the `tile_coord` if it is within the ``grid``,
                  otherwise ``None``.

        >>> grid = TileGrid(SRS(900913))
        >>> grid.limit_tile((-1, 0, 2)) is None
        True
        >>> grid.limit_tile((1, 2, 1)) is None
        True
        >>> grid.limit_tile((1, 2, 2))
        (1, 2, 2)
        """
        x, y, z = tile_coord
        if isinstance(z, str):
            if z not in self.grid_sizes:
                return None
        elif z < 0 or z >= self.levels:
            return None
        grid = self.grid_sizes[z]
        if x < 0 or y < 0 or x >= grid[0] or y >= grid[1]:
            return None
        return x, y, z

    def __repr__(self):
        return '%s(%r, (%.4f, %.4f, %.4f, %.4f),...)' % (
            self.__class__.__name__, self.srs, self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3])

    def is_subset_of(self, other):
        """
        Returns ``True`` if every tile in `self` is present in `other`.
        Tile coordinates might differ and `other` may contain more
        tiles (more levels, larger bbox).
        """
        if self.srs != other.srs:
            return False

        if self.tile_size != other.tile_size:
            return False

        # check if all level tiles from self align with (affected)
        # tiles from other
        for self_level, self_level_res in self.resolutions.iteritems():
            level_size = (
                self.grid_sizes[self_level][0] * self.tile_size[0],
                self.grid_sizes[self_level][1] * self.tile_size[1]
            )
            level_bbox = self._tiles_bbox([
                (0, 0, self_level),
                (self.grid_sizes[self_level][0] - 1, self.grid_sizes[self_level][1] - 1, self_level)
            ])

            try:
                bbox, level = other.get_affected_bbox_and_level(level_bbox, level_size)
            except NoTiles:
                return False
            try:
                bbox, grid_size, tiles = other.get_affected_level_tiles(level_bbox, level)
            except GridError:
                return False

            if other.resolution(level) != self_level_res:
                return False
            if not bbox_equals(bbox, level_bbox):
                return False

        return True


def is_float(x):
    try:
        float(x)
        return True
    except TypeError:
        return False


def tile_grid_to_ogc_tile_matrix_set(tile_grid: TileGrid,
                                     id: str,
                                     title: Optional[str],
                                     uri: Optional[str] = None,
                                     wellKnownScaleSet: Optional[str] = None):

    srs = tile_grid.srs
    if srs.srs_code == "EPSG:4326":
        srs = SRS("OGC:CRS84")
    elif srs.srs_code == "EPSG:900913":
        srs = SRS("EPSG:3857")
    tms = {
        "id": id,
        "title": title,
        "crs": srs.to_ogc_url()
    }
    if uri:
        tms["uri"] = uri
    if wellKnownScaleSet:
        tms["wellKnownScaleSet"] = wellKnownScaleSet

    tileMatrices = []
    for level in range(tile_grid.levels):
        cellSize = tile_grid.resolutions[level]
        cornerOfOrigin = "topLeft" if tile_grid.origin == "ul" else "bottomLeft"
        if tile_grid.origin == "ul":
            pointOfOrigin = [tile_grid.bbox[0], tile_grid.bbox[3]]
        else:
            pointOfOrigin = [tile_grid.bbox[0], tile_grid.bbox[1]]
        if srs.is_axis_order_ne:
            pointOfOrigin = [pointOfOrigin[1], pointOfOrigin[0]]
        tileWidth = tile_grid.tile_size[0]
        tileHeight = tile_grid.tile_size[1]
        matrixWidth = tile_grid.grid_sizes[level][0]
        matrixHeight = tile_grid.grid_sizes[level][1]
        tm = {
            "id": str(level),
            "cellSize": cellSize,
            "scaleDenominator": res_to_ogc_scale(cellSize),
            "cornerOfOrigin": cornerOfOrigin,
            "pointOfOrigin": pointOfOrigin,
            "matrixWidth": matrixWidth,
            "matrixHeight": matrixHeight,
            "tileWidth": tileWidth,
            "tileHeight": tileHeight,
        }
        tileMatrices.append(tm)

    tms["tileMatrices"] = tileMatrices

    return tms



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from mapproxy.client.tile import TileClient, TileURLTemplate
from mapproxy.grid.tile_grid import TileGrid
from mapproxy.srs import SRS
from mapproxy.source.tile import TiledSource
from mapproxy.source.error import HTTPSourceErrorHandler
from mapproxy.layer import MapQuery
from mapproxy.test.http import mock_httpd


TEST_SERVER_ADDRESS = ("127.0.0.1", 56413)
TESTSERVER_URL = ("http://%s:%d" % TEST_SERVER_ADDRESS) + "/%(tms_path)s.png"


class TestTileClientOnError(object):

    def setup_method(self):
        self.grid = TileGrid(SRS(4326), bbox=[-180, -90, 180, 90])
        self.client = TileClient(TileURLTemplate(TESTSERVER_URL))

    def test_cacheable_response(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=True)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                )
            ],
        ):
            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

    def test_image_response(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=False)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                )
            ],
        ):
            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert not resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

    def test_multiple_image_responses(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=False)
        error_handler.add_handler(204, (255, 0, 127, 200), cacheable=True)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                ),
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 204,
                        "headers": {"content-type": "text/plain"},
                    },
                ),
            ],
        ):

            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert not resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 127, 200))]



# -*- coding: utf-8 -*-
# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

from lxml import etree, html

from mapproxy.featureinfo import (
    XSLTransformer,
    TextFeatureInfoDoc,
    XMLFeatureInfoDoc,
    HTMLFeatureInfoDoc,
    JSONFeatureInfoDoc,
    combine_docs,
)
from mapproxy.test.helper import strip_whitespace


class TestXSLTransformer(object):

    def setup_method(self):
        fd, self.xsl_script = tempfile.mkstemp(".xsl")
        os.close(fd)
        xsl = (
            b"""
        <xsl:stylesheet version="1.0"
         xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
         <xsl:template match="/">
            <root>
                <xsl:apply-templates select='/a/b'/>
            </root>
         </xsl:template>
         <xsl:template match="/a/b">
             <foo><xsl:value-of select="text()" /></foo>
         </xsl:template>
        </xsl:stylesheet>""".strip()
        )
        with open(self.xsl_script, "wb") as f:
            f.write(xsl)

    def teardown_method(self):
        os.remove(self.xsl_script)

    def test_transformer(self):
        t = XSLTransformer(self.xsl_script)
        doc = t.transform(XMLFeatureInfoDoc("<a><b>Text</b></a>"))
        assert strip_whitespace(doc.as_string()) == "<root><foo>Text</foo></root>"

    def test_multiple(self):
        t = XSLTransformer(self.xsl_script)
        doc = t.transform(
            XMLFeatureInfoDoc.combine(
                [
                    XMLFeatureInfoDoc(x)
                    for x in [
                        b"<a><b>ab</b></a>",
                        b"<a><b>ab1</b><b>ab2</b><b>ab3</b></a>",
                        b"<a><b>ab1</b><c>ac</c><b>ab2</b></a>",
                    ]
                ]
            )
        )
        assert strip_whitespace(doc.as_string()) == strip_whitespace(
            """
            <root>
              <foo>ab</foo>
              <foo>ab1</foo><foo>ab2</foo><foo>ab3</foo>
              <foo>ab1</foo><foo>ab2</foo>
            </root>
            """
        )
        assert doc.info_type == "xml"


class TestXMLFeatureInfoDocs(object):

    def test_as_string(self):
        input_tree = etree.fromstring("<root></root>")
        doc = XMLFeatureInfoDoc(input_tree)
        assert strip_whitespace(doc.as_string()) == "<root/>"

    def test_as_etree(self):
        doc = XMLFeatureInfoDoc("<root>hello</root>")
        assert doc.as_etree().getroot().text == "hello"

    def test_umlauts(self):
        doc = XMLFeatureInfoDoc('<root>öäüß</root>')
        assert doc.as_etree().getroot().text == 'öäüß'

        input_tree = etree.fromstring(u'<root>öäüß</root>')
        doc = XMLFeatureInfoDoc(input_tree)
        assert doc.as_string() == '<root>öäüß</root>'

    def test_combine(self):
        docs = [
            XMLFeatureInfoDoc("<root><a>foo</a></root>"),
            XMLFeatureInfoDoc("<root><b>bar</b></root>"),
            XMLFeatureInfoDoc("<other_root><a>baz</a></other_root>"),
        ]
        result = XMLFeatureInfoDoc.combine(docs)

        assert strip_whitespace(result.as_string()) == strip_whitespace(
            "<root><a>foo</a><b>bar</b><a>baz</a></root>"
        )
        assert result.info_type == "xml"


class TestXMLFeatureInfoDocsNoLXML(object):

    def setup_method(self):
        from mapproxy import featureinfo

        self.old_etree = featureinfo.etree
        featureinfo.etree = None

    def teardown_method(self):
        from mapproxy import featureinfo

        featureinfo.etree = self.old_etree

    def test_combine(self):
        docs = [
            XMLFeatureInfoDoc(b"<root><a>foo</a></root>"),
            XMLFeatureInfoDoc(b"<root><b>bar</b></root>"),
            XMLFeatureInfoDoc(b"<other_root><a>baz</a></other_root>"),
        ]
        result = XMLFeatureInfoDoc.combine(docs)

        assert (
            "<root><a>foo</a></root>\n<root><b>bar</b></root>\n<other_root><a>baz</a></other_root>"
            == result.as_string()
        )
        assert result.info_type == "text"


class TestHTMLFeatureInfoDocs(object):

    def test_as_string(self):
        input_tree = html.fromstring("<p>Foo")
        doc = HTMLFeatureInfoDoc(input_tree)
        assert "<body><p>Foo</p></body>" in strip_whitespace(doc.as_string())

    def test_as_etree(self):
        doc = HTMLFeatureInfoDoc("<p>hello</p>")
        assert doc.as_etree().find("body/p").text == "hello"

    def test_combine(self):
        docs = [
            HTMLFeatureInfoDoc(b"<html><head><title>Hello</title></head><body><p>baz</p><p>baz2</body></html>"),
            HTMLFeatureInfoDoc(b"<p>foo</p>"),
            HTMLFeatureInfoDoc(b"<body><p>bar</p></body>"),
        ]
        result = HTMLFeatureInfoDoc.combine(docs)
        assert "<title>Hello</title>" in result.as_string()
        assert (
            "<body><p>baz</p><p>baz2</p><p>foo</p><p>bar</p></body>"
            in result.as_string()
        )
        assert result.info_type == "html"

    def test_combine_parts(self):
        docs = [
            HTMLFeatureInfoDoc("<p>foo</p>"),
            HTMLFeatureInfoDoc("<body><p>bar</p></body>"),
            HTMLFeatureInfoDoc("<html><head><title>Hello</title></head><body><p>baz</p><p>baz2</body></html>"),
        ]
        result = HTMLFeatureInfoDoc.combine(docs)

        assert (
            "<body><p>foo</p><p>bar</p><p>baz</p><p>baz2</p></body>"
            in result.as_string()
        )
        assert result.info_type == "html"


class TestHTMLFeatureInfoDocsNoLXML(object):

    def setup_method(self):
        from mapproxy import featureinfo

        self.old_etree = featureinfo.etree
        featureinfo.etree = None

    def teardown_method(self):
        from mapproxy import featureinfo

        featureinfo.etree = self.old_etree

    def test_combine(self):
        docs = [
            HTMLFeatureInfoDoc(b"<html><head><title>Hello<body><p>baz</p><p>baz2"),
            HTMLFeatureInfoDoc(b"<p>foo</p>"),
            HTMLFeatureInfoDoc(b"<body><p>bar</p></body>"),
        ]
        result = HTMLFeatureInfoDoc.combine(docs)

        assert (
            "<html><head><title>Hello<body><p>baz</p>"
            "<p>baz2\n<p>foo</p>\n<body><p>bar</p></body>" == result.as_string()
        )
        assert result.info_type == "text"


class TestJSONFeatureInfoDocs(object):

    def test_combine(self):
        docs = [
            JSONFeatureInfoDoc("{}"),
            JSONFeatureInfoDoc('{"results": [{"foo": 1}]}'),
            JSONFeatureInfoDoc('{"results": [{"bar": 2}]}'),
        ]
        result = JSONFeatureInfoDoc.combine(docs)

        assert """{"results": [{"foo": 1}, {"bar": 2}]}""" == result.as_string()
        assert result.info_type == "json"


class TestCombineDocs(object):

    def test_combine_json(self):
        docs = [
            JSONFeatureInfoDoc("{}"),
            JSONFeatureInfoDoc('{"results": [{"foo": 1}]}'),
            JSONFeatureInfoDoc('{"results": [{"bar": 2}]}'),
            JSONFeatureInfoDoc("{}"),
        ]
        result, infotype = combine_docs(docs, None)

        assert """{"results": [{"foo": 1}, {"bar": 2}]}""" == result
        assert infotype == "json"

    def test_combine_json_with_null_response(self):
        docs = [
            JSONFeatureInfoDoc("null"),
            JSONFeatureInfoDoc('{"results": [{"foo": 1}]}'),
            JSONFeatureInfoDoc("null"),
        ]
        result, infotype = combine_docs(docs, None)

        assert """{"results": [{"foo": 1}]}""" == result
        assert infotype == "json"

    def test_combine_json_with_only_empty_responses(self):
        docs = [
            JSONFeatureInfoDoc("{}"),
            JSONFeatureInfoDoc("{}"),
        ]
        result, infotype = combine_docs(docs, None)

        assert """{}""" == result
        assert infotype == "json"

    def test_combine_xml(self):
        docs = [
            XMLFeatureInfoDoc("<root><a>foo</a></root>"),
            XMLFeatureInfoDoc("<root><b>bar</b></root>"),
        ]
        result, infotype = combine_docs(docs, None)

        assert """<root><a>foo</a><b>bar</b></root>""" == result
        assert infotype == "xml"

    def test_combine_transform_xml(self, tmpdir):
        xsl = tmpdir.join("transform.xsl")
        xsl.write(
            """<xsl:stylesheet version="1.0"
         xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
         <xsl:template match="/">
            <root>
                <xsl:apply-templates select='/a/b'/>
            </root>
         </xsl:template>
         <xsl:template match="/a/b">
             <foo><xsl:value-of select="text()" /></foo>
         </xsl:template>
        </xsl:stylesheet>
        """
        )
        transf = XSLTransformer(xsl.strpath)
        docs = [
            XMLFeatureInfoDoc("<a><b>foo1</b></a>"),
            XMLFeatureInfoDoc("<a><b>foo2</b></a>"),
        ]
        result, infotype = combine_docs(docs, transf)

        assert """<root><foo>foo1</foo><foo>foo2</foo></root>""" == result
        assert infotype is None

    def test_combine_mixed(self):
        docs = [
            JSONFeatureInfoDoc(b'{"results": [{"foo": 1}]}'),
            TextFeatureInfoDoc(b"Plain Text"),
        ]
        result, infotype = combine_docs(docs, None)

        assert """{"results": [{"foo": 1}]}\nPlain Text""" == result
        assert infotype == "text"



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

F_JSON = "json"
F_HTML = "html"
F_PNG = "png"
F_JPEG = "jpeg"

#: Formats allowed for ?f= requests (order matters for complex MIME types)
FORMAT_TYPES = OrderedDict(
    (
        (F_HTML, "text/html"),
        (F_JSON, "application/json"),
        (F_PNG, "image/png"),
        (F_JPEG, "image/jpeg"),
    )
)

MEDIA_TYPE_OPENAPI_3_0 = "application/vnd.oai.openapi+json;version=3.0"



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import json
import codecs
from functools import partial
from contextlib import closing

import shapely.wkt
import shapely.geometry
import shapely.ops
import shapely.prepared
from shapely.errors import ShapelyError

from mapproxy.grid.tile_grid import tile_grid

import logging
log_config = logging.getLogger('mapproxy.config.coverage')


class GeometryError(Exception):
    pass


class EmptyGeometryError(Exception):
    pass


class CoverageReadError(Exception):
    pass


def load_datasource(datasource, where=None):
    """
    Loads polygons from WKT text files or OGR datasources.

    Returns a list of Shapely Polygons.
    """
    # check if it is a  wkt or geojson file
    if os.path.exists(os.path.abspath(datasource)):
        with open(os.path.abspath(datasource), 'rb') as fp:
            data = fp.read(50)
        if data.lower().lstrip().startswith((b'polygon', b'multipolygon')):
            return load_polygons(datasource)
        # only load geojson directly if we don't have a filter
        if where is None and data and data.startswith(b'{'):
            return load_geojson(datasource)
    # otherwise pass to OGR
    return load_ogr_datasource(datasource, where=where)


def load_ogr_datasource(datasource, where=None):
    """
    Loads polygons from any OGR datasource.

    Returns a list of Shapely Polygons.
    """
    from mapproxy.util.ogr import OGRShapeReader, OGRShapeReaderError

    polygons = []
    try:
        with closing(OGRShapeReader(datasource)) as reader:
            for wkt in reader.wkts(where):
                if not isinstance(wkt, str):
                    wkt = wkt.decode()
                try:
                    geom = shapely.wkt.loads(wkt)
                except ShapelyError as ex:
                    raise GeometryError(ex)
                if geom.geom_type == 'Polygon':
                    polygons.append(geom)
                elif geom.geom_type == 'MultiPolygon':
                    for p in geom.geoms:
                        polygons.append(p)
                else:
                    log_config.warning('skipping %s geometry from %s: not a Polygon/MultiPolygon',
                                       geom.geom_type, datasource)
    except OGRShapeReaderError as ex:
        raise CoverageReadError(ex)

    return polygons


def load_polygons(geom_files):
    """
    Loads WKT polygons from one or more text files.

    Returns a list of Shapely Polygons.
    """
    polygons = []
    if isinstance(geom_files, str):
        geom_files = [geom_files]

    for geom_file in geom_files:
        # open with utf-8-sig encoding to get rid of UTF8 BOM from MS Notepad
        with codecs.open(geom_file, encoding='utf-8-sig') as f:
            polygons.extend(load_polygon_lines(f, source=geom_files))

    return polygons


def load_geojson(datasource):
    with open(datasource) as f:
        geojson = json.load(f)
        t = geojson.get('type')
        if not t:
            raise CoverageReadError("not a GeoJSON")
        geometries = []
        if t == 'FeatureCollection':
            for f in geojson.get('features'):
                geom = f.get('geometry')
                if geom:
                    geometries.append(geom)
        elif t == 'Feature':
            if 'geometry' in geojson:
                geometries.append(geojson['geometry'])
        elif t in ('Polygon', 'MultiPolygon'):
            geometries.append(geojson)
        else:
            log_config.warning('skipping feature of type %s from %s: not a Polygon/MultiPolygon',
                               t, datasource)

    polygons = []
    for geom in geometries:
        geom = shapely.geometry.shape(geom)
        if geom.geom_type == 'Polygon':
            polygons.append(geom)
        elif geom.geom_type == 'MultiPolygon':
            for p in geom.geoms:
                polygons.append(p)
        else:
            log_config.warning('ignoring non-polygon geometry (%s) from %s',
                               geom.geom_type, datasource)

    return polygons


def load_polygon_lines(line_iter, source='<string>'):
    polygons = []
    for line in line_iter:
        if not line.strip():
            continue
        geom = shapely.wkt.loads(line)
        if geom.geom_type == 'Polygon':
            polygons.append(geom)
        elif geom.geom_type == 'MultiPolygon':
            for p in geom.geoms:
                polygons.append(p)
        else:
            log_config.warning('ignoring non-polygon geometry (%s) from %s',
                               geom.geom_type, source)

    return polygons


def build_multipolygon(polygons, simplify=False):
    if not polygons:
        raise EmptyGeometryError('no polygons')

    if len(polygons) == 1:
        geom = polygons[0]
        if simplify:
            geom = simplify_geom(geom)
        return geom.bounds, geom

    if simplify:
        polygons = [simplify_geom(g) for g in polygons]

    # eliminate any self-overlaps
    mp = shapely.ops.unary_union(polygons)

    return mp.bounds, mp


def simplify_geom(geom):
    bounds = geom.bounds
    if not bounds:
        raise EmptyGeometryError('Empty geometry given')
    w, h = bounds[2] - bounds[0], bounds[3] - bounds[1]
    tolerance = min((w/1e6, h/1e6))
    geom = geom.simplify(tolerance, preserve_topology=True)
    if not geom.is_valid:
        geom = geom.buffer(0)
    return geom


def bbox_polygon(bbox):
    """
    Create Polygon that covers the given bbox.
    """
    return shapely.geometry.Polygon((
        (bbox[0], bbox[1]),
        (bbox[2], bbox[1]),
        (bbox[2], bbox[3]),
        (bbox[0], bbox[3]),
    ))


def transform_geometry(from_srs, to_srs, geometry):
    transf = partial(transform_xy, from_srs, to_srs)

    if geometry.geom_type == 'Polygon':
        result = transform_polygon(transf, geometry)
    elif geometry.geom_type == 'MultiPolygon':
        result = transform_multipolygon(transf, geometry)
    else:
        raise ValueError('cannot transform %s' % geometry.geom_type)

    if not result.is_valid:
        result = result.buffer(0)
    return result


def transform_polygon(transf, polygon):
    ext = transf(polygon.exterior.xy)
    ints = [transf(ring.xy) for ring in polygon.interiors]
    return shapely.geometry.Polygon(ext, ints)


def transform_multipolygon(transf, multipolygon):
    transformed_polygons = []
    for polygon in multipolygon.geoms:
        transformed_polygons.append(transform_polygon(transf, polygon))
    return shapely.geometry.MultiPolygon(transformed_polygons)


def transform_xy(from_srs, to_srs, xy):
    return list(from_srs.transform_to(to_srs, list(zip(*xy))))


def flatten_to_polygons(geometry):
    """
    Return a list of all polygons of this (multi)`geometry`.
    """
    if geometry.geom_type == 'Polygon':
        return [geometry]

    if geometry.geom_type == 'MultiPolygon':
        return list(geometry.geoms)

    if hasattr(geometry, 'geoms'):
        # GeometryCollection or MultiLineString? return list of all polygons
        geoms = []
        for part in geometry.geoms:
            if part.type == 'Polygon':
                geoms.append(part)

        if geoms:
            return geoms

    return []


def load_expire_tiles(expire_dir, grid=None):
    if grid is None:
        grid = tile_grid(3857, origin='nw')
    tiles = set()

    def parse(filename):
        with open(filename) as f:
            try:
                for line in f:
                    if not line:
                        continue
                    tile = tuple(map(int, line.split('/')))
                    tiles.add(tile)
            except Exception:
                log_config.warning('found error in %s, skipping rest of file', filename)

    if os.path.isdir(expire_dir):
        for root, dirs, files in os.walk(expire_dir):
            for name in files:
                filename = os.path.join(root, name)
                parse(filename)
    else:
        parse(expire_dir)

    boxes = []
    for tile in tiles:
        z, x, y = tile
        boxes.append(shapely.geometry.box(*grid.tile_bbox((x, y, z))))

    return boxes



# This file is part of the MapProxy project.
# Copyright (C) 2010-2016 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Image and tile manipulation (transforming, merging, etc).
"""
import json
from collections import namedtuple
from PIL import Image, ImageColor, ImageChops, ImageMath
from mapproxy.image import BlankImageSource, ImageSource
from mapproxy.image.opts import create_image, ImageOptions
from mapproxy.image.mask import mask_image

import logging
log = logging.getLogger('mapproxy.image')


class LayerMerger(object):
    """
    Merge multiple layers into one image.
    """

    def __init__(self):
        self.layers = []
        self.cacheable = True

    def add(self, img, coverage=None):
        """
        Add one layer image to merge. Bottom-layers first.
        """
        if img is not None:
            self.layers.append((img, coverage))


class LayerMerger(LayerMerger):

    def merge(self, image_opts, size=None, bbox=None, bbox_srs=None, coverage=None):
        """
        Merge the layers. If the format is not 'png' just return the last image.

        :param format: The image format for the result.
        :param size: The size for the merged output.
        :rtype: `ImageSource`
        """
        if not self.layers:
            return BlankImageSource(size=size, image_opts=image_opts, cacheable=True)
        if len(self.layers) == 1:
            layer_img, layer_coverage = self.layers[0]
            layer_opts = layer_img.image_opts
            if (((layer_opts and not layer_opts.transparent) or image_opts.transparent)
                and (not size or size == layer_img.size)
                and (not layer_coverage or not layer_coverage.clip)
                    and not coverage):
                # layer is opaque, no need to make transparent or add bgcolor
                return layer_img

        if size is None:
            size = self.layers[0][0].size

        cacheable = self.cacheable
        result = create_image(size, image_opts)
        for layer_img, layer_coverage in self.layers:
            if not layer_img.cacheable:
                cacheable = False
            img = layer_img.as_image()
            layer_image_opts = layer_img.image_opts
            if layer_image_opts is None:
                opacity = None
            else:
                opacity = layer_image_opts.opacity

            if layer_coverage and layer_coverage.clip:
                img = mask_image(img, bbox, bbox_srs, layer_coverage)

            merge_composite = result.mode == 'RGBA'

            if 'transparency' in img.info:
                # non-paletted PNGs can have a fixed transparency value
                # convert to RGBA to have full alpha
                img = img.convert('RGBA')

            if merge_composite:
                if opacity is not None and opacity < 1.0:
                    # fade-out img to add opacity value
                    img = img.convert("RGBA")
                    alpha = img.split()[3]
                    alpha = ImageChops.multiply(
                        alpha,
                        ImageChops.constant(alpha, int(255 * opacity))
                    )
                    img.putalpha(alpha)
                if img.mode in ('RGBA', 'P'):
                    # assume paletted images have transparency
                    if img.mode == 'P':
                        img = img.convert('RGBA')
                    result = Image.alpha_composite(result, img)
                else:
                    result.paste(img, (0, 0))
            else:
                if opacity is not None and opacity < 1.0:
                    img = img.convert(result.mode)
                    result = Image.blend(result, img, layer_image_opts.opacity)
                elif img.mode in ('RGBA', 'P'):
                    # assume paletted images have transparency
                    if img.mode == 'P':
                        img = img.convert('RGBA')
                    # paste w transparency mask from layer
                    result.paste(img, (0, 0), img)
                else:
                    result.paste(img, (0, 0))

        # apply global clip coverage
        if coverage:
            bg = create_image(size, image_opts)
            mask = mask_image(result, bbox, bbox_srs, coverage)
            bg.paste(result, (0, 0), mask)
            result = bg

        return ImageSource(result, size=size, image_opts=image_opts, cacheable=cacheable)


band_ops = namedtuple("band_ops", ["dst_band", "src_img", "src_band", "factor"])


class BandMerger(object):
    """
    Merge bands from multiple sources into one image.

       sources:
           r: [{source: nir_cache, band: 0, factor: 0.4}, {source: dop_cache, band: 0, factor: 0.6}]
           g: [{source: dop_cache, band: 2}]
           b: [{source: dop_cache, band: 1}]

       sources:
           l: [
               {source: dop_cache, band: 0, factor: 0.6},
               {source: dop_cache, band: 1, factor: 0.3},
               {source: dop_cache, band: 2, factor: 0.1},
           ]
    """

    def __init__(self, mode=None):
        self.ops = []
        self.cacheable = True
        self.mode = mode
        self.max_band = {}
        self.max_src_images = 0

    def add_ops(self, dst_band, src_img, src_band, factor=1.0):
        self.ops.append(band_ops(
            dst_band=dst_band,
            src_img=src_img,
            src_band=src_band,
            factor=factor,
        ))
        # store highest requested band index for each source
        self.max_band[src_img] = max(self.max_band.get(src_img, 0), src_band)
        self.max_src_images = max(src_img+1, self.max_src_images)

    def merge(self, sources, image_opts, size=None, bbox=None, bbox_srs=None, coverage=None):
        if len(sources) < self.max_src_images:
            return BlankImageSource(size=size, image_opts=image_opts, cacheable=True)

        if size is None:
            size = sources[0].size

        # load src bands
        src_img_bands = []
        for i, layer_img in enumerate(sources):
            img = layer_img.as_image()

            if i not in self.max_band:
                # do not split img if not requested by any op
                src_img_bands.append(None)
                continue

            if self.max_band[i] == 3 and img.mode != 'RGBA':
                # convert to RGBA if band idx 3 is requestd (e.g. P or RGB src)
                img = img.convert('RGBA')
            elif img.mode == 'P':
                img = img.convert('RGB')
            src_img_bands.append(img.split())

        tmp_mode = self.mode

        if tmp_mode == 'RGBA':
            result_bands = [None, None, None, None]
        elif tmp_mode == 'RGB':
            result_bands = [None, None, None]
        elif tmp_mode == 'L':
            result_bands = [None]
        else:
            raise ValueError("unsupported destination mode %s", image_opts.mode)

        for op in self.ops:
            chan = src_img_bands[op.src_img][op.src_band]
            if op.factor != 1.0:
                # eval is deprecated since Pillow==10.3.0, lambda_eval is the replacement
                if hasattr(ImageMath, 'lambda_eval'):
                    chan = ImageMath.lambda_eval(
                        lambda args: args["convert"](args["int"](args["float"](args["a"]) * op.factor), 'L'),
                        a=chan
                    )
                else:
                    chan = ImageMath.eval("convert(int(float(a) * %f), 'L')" % op.factor, a=chan)
                if result_bands[op.dst_band] is None:
                    result_bands[op.dst_band] = chan
                else:
                    result_bands[op.dst_band] = ImageChops.add(
                        result_bands[op.dst_band],
                        chan,
                    )
            else:
                result_bands[op.dst_band] = chan

        for i, b in enumerate(result_bands):
            if b is None:
                # band not set
                b = Image.new("L", size, 255 if i == 3 else 0)
                result_bands[i] = b

        result = Image.merge(tmp_mode, result_bands)
        return ImageSource(result, size=size, image_opts=image_opts)


def merge_images(layers, image_opts, size=None, bbox=None, bbox_srs=None, merger=None):
    """
    Merge multiple images into one.

    :param images: list of `ImageSource`, bottom image first
    :param format: the format of the output `ImageSource`
    :param size: size of the merged image, if ``None`` the size
                 of the first image is used
    :param bbox: Bounding box
    :param bbox_srs: Bounding box SRS
    :param merger: Image merger
    :rtype: `ImageSource`
    """
    if merger is None:
        merger = LayerMerger()

    # BandMerger does not have coverage support, passing only images
    if isinstance(merger, BandMerger):
        sources = [x[0] if isinstance(x, tuple) else x for x in layers]
        return merger.merge(sources, image_opts=image_opts, size=size, bbox=bbox, bbox_srs=bbox_srs)

    for layer in layers:
        if isinstance(layer, tuple):
            merger.add(layer[0], layer[1])
        else:
            merger.add(layer)

    return merger.merge(image_opts=image_opts, size=size, bbox=bbox, bbox_srs=bbox_srs)


def concat_legends(legends, format='png', size=None, bgcolor='#ffffff', transparent=True):
    """
    Merge multiple legends into one
    :param images: list of `ImageSource`, bottom image first
    :param format: the format of the output `ImageSource`
    :param size: size of the merged image, if ``None`` the size
                 will be calculated
    :rtype: `ImageSource`
    """
    if not legends:
        return BlankImageSource(size=(1, 1), image_opts=ImageOptions(bgcolor=bgcolor, transparent=transparent))
    if len(legends) == 1:
        return legends[0]

    legends = legends[:]
    legends.reverse()
    if size is None:
        legend_width = 0
        legend_height = 0
        legend_position_y = []
        # iterate through all legends, last to first, calc img size and remember the y-position
        for legend in legends:
            legend_position_y.append(legend_height)
            tmp_img = legend.as_image()
            legend_width = max(legend_width, tmp_img.size[0])
            legend_height += tmp_img.size[1]  # images shall not overlap themselfs

        size = [legend_width, legend_height]
    bgcolor = ImageColor.getrgb(bgcolor)

    if transparent:
        img = Image.new('RGBA', size, bgcolor+(0,))
    else:
        img = Image.new('RGB', size, bgcolor)
    for i in range(len(legends)):
        legend_img = legends[i].as_image()
        if legend_img.mode == 'RGBA':
            # paste w transparency mask from layer
            img.paste(legend_img, (0, legend_position_y[i]), legend_img)
        else:
            img.paste(legend_img, (0, legend_position_y[i]))
    return ImageSource(img, image_opts=ImageOptions(format=format))


def concat_json_legends(legends):
    """
    Merge multiple json legends into one
    :param legends: list of HttpResponse objects
    :rtype json formatted string
    """
    legends_dict = {'Legend': []}
    legends = legends[:]
    for legend in legends:
        legend_str = legend.read().decode('utf-8')
        legend_dict = json.loads(legend_str)
        if 'Legend' in legend_dict:
            legend_dict = legend_dict['Legend'][0]
        legends_dict['Legend'].append(legend_dict)
    return json.dumps(legends_dict)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import yaml


class YAMLError(Exception):
    pass


def load_yaml_file(file_or_filename):
    """
    Load yaml from file object or filename.
    """
    if isinstance(file_or_filename, str):
        with open(file_or_filename, 'rb') as f:
            return load_yaml(f)
    return load_yaml(file_or_filename)


def _load_yaml(doc):
    # try different methods to load yaml
    try:
        if getattr(yaml, '__with_libyaml__', False):
            try:
                return yaml.load(doc, Loader=yaml.CSafeLoader)
            except AttributeError:
                # handle cases where __with_libyaml__ is True but
                # CLoader doesn't work (missing .dispose())
                return yaml.safe_load(doc)
        return yaml.safe_load(doc)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as ex:
        raise YAMLError(str(ex))


def load_yaml(doc):
    """
    Load yaml from file object or string.
    """
    data = _load_yaml(doc)
    if type(data) is not dict:
        # all configs are dicts, raise YAMLError to prevent later AttributeErrors (#352)
        raise YAMLError("configuration not a YAML dictionary")
    return data



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging

import sys
import time
import threading
import multiprocessing
from io import BytesIO

from mapproxy.grid.tile_grid import tile_grid
from mapproxy.image import ImageSource
from mapproxy.image.opts import ImageOptions
from mapproxy.layer import MapExtent, DefaultMapExtent, BlankImage, MapLayer
from mapproxy.source import SourceError
from mapproxy.client.log import log_request
from mapproxy.util.py import reraise_exception
from mapproxy.util.async_ import run_non_blocking

try:
    import mapnik
    mapnik
except ImportError:
    try:
        # for 2.0 alpha/rcs and first 2.0 release
        import mapnik2 as mapnik
    except ImportError:
        mapnik = None

import queue
Queue = queue.Queue
Empty = queue.Empty
Full = queue.Full
MAX_UNUSED_MAPS = 10

# fake 2.0 API for older versions
if mapnik and not hasattr(mapnik, 'Box2d'):
    mapnik.Box2d = mapnik.Envelope

log = logging.getLogger(__package__)


class MapnikSource(MapLayer):
    supports_meta_tiles = True

    def __init__(self, mapfile, layers=None, image_opts=None, coverage=None,
                 res_range=None, lock=None, reuse_map_objects=False,
                 scale_factor=None, multithreaded=False):
        MapLayer.__init__(self, image_opts=image_opts)
        self.mapfile = mapfile
        self.coverage = coverage
        self.res_range = res_range
        self.layers = set(layers) if layers else None
        self.scale_factor = scale_factor
        self.lock = lock
        self.multithreaded = multithreaded
        self._cache_map_obj = reuse_map_objects
        if self.coverage:
            self.extent = MapExtent(self.coverage.bbox, self.coverage.srs)
        else:
            self.extent = DefaultMapExtent()
        if multithreaded:
            # global objects allow caching over multiple instances within the same worker process
            global _map_objs  # mapnik map objects by cachekey
            _map_objs = {}
            global _map_objs_lock
            _map_objs_lock = threading.Lock()
            global _map_objs_queues  # queues of unused mapnik map objects by PID and mapfile
            _map_objs_queues = {}
        else:
            # instance variables guarantee separation of caches
            self._map_objs = {}
            self._map_objs_lock = threading.Lock()

    def _map_cache(self):
        """Get the cache for map objects.

        Uses an instance variable for containment when multithreaded
        operation is disabled.
        """
        if self.multithreaded:
            return _map_objs
        return self._map_objs

    def _map_cache_lock(self):
        """Get the cache-locks for map objects.

        Uses an instance variable for containment when multithreaded
        operation is disabled.
        """
        if self.multithreaded:
            return _map_objs_lock
        return self._map_objs_lock

    def get_map(self, query):
        if self.res_range and not self.res_range.contains(query.bbox, query.size,
                                                          query.srs):
            raise BlankImage()
        if self.coverage and not self.coverage.intersects(query.bbox, query.srs):
            raise BlankImage()

        try:
            resp = self.render(query)
        except RuntimeError as ex:
            log.error('could not render Mapnik map: %s', ex)
            reraise_exception(SourceError(ex.args[0]), sys.exc_info())
        resp.opacity = self.opacity
        return resp

    def render(self, query):
        mapfile = self.mapfile
        if '%(webmercator_level)' in mapfile:
            _bbox, level = tile_grid(3857).get_affected_bbox_and_level(
                query.bbox, query.size, req_srs=query.srs)
            mapfile = mapfile % {'webmercator_level': level}

        if self.lock:
            with self.lock():
                return self.render_mapfile(mapfile, query)
        else:
            return self.render_mapfile(mapfile, query)

    def _create_map_obj(self, mapfile, process_id):
        m = mapnik.Map(0, 0)
        mapnik.load_map(m, str(mapfile))
        m.map_obj_pid = process_id
        return m

    def _get_map_obj(self, mapfile):
        if not self.multithreaded:
            return self._create_map_obj(mapfile, None)

        process_id = multiprocessing.current_process()._identity
        queue_cachekey = (process_id, mapfile)
        if queue_cachekey in _map_objs_queues:
            try:
                m = _map_objs_queues[queue_cachekey].get_nowait()
                # check explicitly for the process ID to ensure that
                # map objects cannot move between processes
                if m.map_obj_pid == process_id:
                    return m
            except Empty:
                pass
        return self._create_map_obj(mapfile, process_id)

    def _put_unused_map_obj(self, mapfile, m):
        process_id = multiprocessing.current_process()._identity
        queue_cachekey = (process_id, mapfile)
        if queue_cachekey not in _map_objs_queues:
            _map_objs_queues[queue_cachekey] = Queue(MAX_UNUSED_MAPS)
        try:
            _map_objs_queues[queue_cachekey].put_nowait(m)
        except Full:
            # cleanup the data and drop the map, so it can be garbage collected
            m.remove_all()
            mapnik.clear_cache()

    def _get_cachekey(self, mapfile):
        if not self.multithreaded or self._cache_map_obj:
            # all MapnikSources with the same mapfile share the same Mapnik Map.
            return (None, None, mapfile)
        thread_id = threading.current_thread().ident
        process_id = multiprocessing.current_process()._identity
        return (process_id, thread_id, mapfile)

    def _cleanup_unused_cached_maps(self, mapfile):
        if not self.multithreaded:
            return
        # clean up no longer used cached maps
        process_id = multiprocessing.current_process()._identity
        process_cache_keys = [k for k in self._map_cache().keys()
                              if k[0] == process_id]
        # To avoid time-consuming cleanup whenever one thread in the
        # threadpool finishes, allow ignoring up to 5 dead mapnik
        # instances.  (5 is empirical)
        if len(process_cache_keys) > (5 + threading.active_count()):
            active_thread_ids = set(i.ident for i in threading.enumerate())
            for k in process_cache_keys:
                with self._map_cache_lock():
                    if not k[1] in active_thread_ids and k in self._map_cache():
                        try:
                            m = self._map_cache()[k]
                            del self._map_cache()[k]
                            # put the map into the queue of unused
                            # maps so it can be re-used from another
                            # thread.
                            self._put_unused_map_obj(mapfile, m)
                        except KeyError:
                            continue

    def map_obj(self, mapfile):
        # cache loaded map objects
        # only works when a single proc/thread accesses the map
        # (forking the render process doesn't work because of open database
        #  file handles that gets passed to the child)
        # segment the cache by process and thread to avoid interference
        cachekey = self._get_cachekey(mapfile)
        with self._map_cache_lock():
            if cachekey not in self._map_cache():
                self._map_cache()[cachekey] = self._get_map_obj(mapfile)
            mapnik_map = self._map_cache()[cachekey]

        self._cleanup_unused_cached_maps(mapfile)

        return mapnik_map

    def render_mapfile(self, mapfile, query):
        return run_non_blocking(self._render_mapfile, (mapfile, query))

    def _render_mapfile(self, mapfile, query):
        start_time = time.time()

        m = self.map_obj(mapfile)
        m.resize(query.size[0], query.size[1])
        m.srs = '+init=%s' % str(query.srs.srs_code.lower())
        envelope = mapnik.Box2d(*query.bbox)
        m.zoom_to_box(envelope)
        data = None

        try:
            if self.layers:
                i = 0
                for layer in m.layers[:]:
                    if layer.name != 'Unknown' and layer.name not in self.layers:
                        del m.layers[i]
                    else:
                        i += 1

            img = mapnik.Image(query.size[0], query.size[1])
            if self.scale_factor:
                mapnik.render(m, img, self.scale_factor)
            else:
                mapnik.render(m, img)
            data = img.tostring(str(query.format))
        finally:
            size = None
            if data:
                size = len(data)
            log_request('%s:%s:%s:%s' % (mapfile, query.bbox, query.srs.srs_code, query.size),
                        status='200' if data else '500', size=size, method='API', duration=time.time()-start_time)

        return ImageSource(BytesIO(data), size=query.size,
                           image_opts=ImageOptions(format=query.format))



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

server = ['wms', 'tms', 'kml']

wms = dict(
    image_formats=['image/png', 'image/jpeg', 'image/gif', 'image/GeoTIFF', 'image/tiff'],
    srs=set(['EPSG:4326', 'EPSG:4258', 'CRS:84', 'EPSG:900913', 'EPSG:3857']),
    strict=False,
    request_parser='default',
    client_request='default',
    concurrent_layer_renderer=1,
    max_output_pixels=4000*4000,
)

# Global settings for OGCAPI Maps and Tiles service
ogcapi = dict(
    image_formats=['image/png', 'image/jpeg'],
    map_srs=set(['EPSG:4326', 'EPSG:3857']),
    concurrent_layer_renderer=1,
    max_output_pixels=4000*4000,
)

debug_mode = False

background = dict(
    url="https://tile.openstreetmap.org/{z}/{x}/{y}.png"
)

srs = dict(
    # user sets
    axis_order_ne=set(),
    axis_order_en=set(),
    # default sets, both will be combined in config:load_base_config
    axis_order_ne_=set(['EPSG:4326', 'EPSG:4258', 'EPSG:31466', 'EPSG:31467', 'EPSG:31468']),
    axis_order_en_=set(['CRS:84', 'EPSG:900913', 'EPSG:25831', 'EPSG:25832', 'EPSG:25833']),
)

image = dict(
    # nearest, bilinear, bicubic
    resampling_method='bicubic',
    jpeg_quality=90,
    stretch_factor=1.15,
    max_shrink_factor=4.0,
    paletted=True,
    transparent_color_tolerance=5,
    font_dir=None,
)
# number of concurrent requests to a tile source


services_conf = 'services.yaml'
log_conf = 'log.ini'

# directory with mapproxy/service/templates/* files
template_dir = None

cache = dict(
    base_dir='./cache_data',
    lock_dir='./cache_data/tile_locks',
    max_tile_limit=500,
    concurrent_tile_creators=2,
    meta_size=(4, 4),
    meta_buffer=80,
    minimize_meta_requests=False,
    link_single_color_images=False,
    sqlite_timeout=30,
)

grid = dict(
    tile_size=(256, 256),
)

grids = dict(
    GLOBAL_GEODETIC=dict(
        srs='EPSG:4326', origin='sw', name='GLOBAL_GEODETIC'
    ),
    GLOBAL_MERCATOR=dict(
        srs='EPSG:900913', origin='sw', name='GLOBAL_MERCATOR'
    ),
    GLOBAL_WEBMERCATOR=dict(
        srs='EPSG:3857', origin='nw', name='GLOBAL_WEBMERCATOR'
    )
)

tiles = dict(
    expires_hours=72,
)

http = dict(
    ssl_ca_certs=None,
    ssl_no_cert_checks=False,
    client_timeout=60,
    concurrent_requests=0,
    method='AUTO',
    access_control_allow_origin='*',
    hide_error_details=True,
    manage_cookies=False,
)
